from .basic import (ee_mean, ee_mean_variance, ee_mean_robust, ee_mean_geometric,
                    ee_percentile, ee_positive_mean_deviation)

from .causal import (ee_ipw, ee_ipw_msm, ee_gformula, ee_aipw, ee_gestimation_snmm,
                     ee_iv_causal, ee_2sls,
                     ee_mean_sensitivity_analysis)

from .pharmacokinetics import (ee_emax, ee_emax_ed,
                               ee_loglogistic, ee_loglogistic_ed)

from .measurement import (ee_rogan_gladen, ee_rogan_gladen_extended, ee_regression_calibration
                          )

from .regression import (ee_regression, ee_glm, ee_mlogit,
                         ee_robust_regression,
                         ee_ridge_regression, ee_lasso_regression, ee_dlasso_regression,
                         ee_elasticnet_regression, ee_bridge_regression,
                         ee_additive_regression)

from .survival import (ee_survival_model, ee_exponential_model, ee_weibull_model,
                       ee_exponential_measure, ee_weibull_measure,
                       ee_aft_weibull, ee_aft_weibull_measure,
                       ee_aft)



import numpy as np


def convert_survival_measures(survival, hazard, measure):
    r"""Function to convert between different survival analysis metrics. This function takes input survival and hazard
    values for a specific time (or for a vector of times) and converts them to the desired metric. Let :math:`S(t)` be
    survival and :math:`h(t)` be the hazard at time :math:`t`. The different metrics are then obtained via

    .. math::

        S(t) = S(t)              \\
        F(t) = 1 - S(t)          \\
        H(t) = - \log(S(t))      \\
        h(t) = h(t)              \\
        f(t) = h(t) \times S(t)

    where :math:`F(t)` is the risk or cumulative distribution function, :math:`H(t)` is the cumulative hazard, and
    :math:`f(t)` is the density function.

    This functionality is principally intended to be used internally to manage conversions between

    Parameters
    ----------
    survival : float, ndarray, list, vector
        Survival at a designated time or times.
    hazard : float, ndarray, list, vector
        Hazard at a designated time or times.
    measure : str
        Survival analysis metric to return.

    Returns
    -------
    array :
        Returns a float or NumPy array of the transformed metric(s) depending on the input types.

    Examples
    --------
    The following illustrates use of ``convert_survival_measures`` to convert between measures

    >>> from delicatessen.helper import convert_survival_measures

    Different measures can be computed via

    >>> convert_survival_measures(survival=0.75, hazard=0.006, measure='survival')
    >>> convert_survival_measures(survival=0.75, hazard=0.006, measure='risk')
    >>> convert_survival_measures(survival=0.75, hazard=0.006, measure='density')
    >>> convert_survival_measures(survival=0.75, hazard=0.006, measure='hazard')
    >>> convert_survival_measures(survival=0.75, hazard=0.006, measure='cumulative_hazard')

    Inputs can also be vectors of matching lengths

    >>> survival_t = [0.75, 0.50, 0.40]
    >>> hazard_t = [0.006, 0.004, 0.001]
    >>> convert_survival_measures(survival=survival_t, hazard=hazard_t, measure='survival')

    References
    ----------
    Collett D. (2015). Survival analysis. In: Modelling Survival Data in Medical Research. CRC press. pg 10-14
    """
    survival = np.asarray(survival)
    hazard = np.asarray(hazard)

    # Converting survival or hazard to desired measure
    if measure == "survival":
        metric = survival  # S(t) = S(t)
    elif measure in ["risk", "cdf", "cumulative_distribution_function"]:
        metric = 1 - survival  # F(t) = 1 - S(t)
    elif measure in ["cumulative_hazard", "chazard"]:
        metric = -1 * np.log(survival)  # H(t) = -log(S(t))
    elif measure == "hazard":
        metric = hazard  # h(t) = h(t)
    elif measure in ["density", "pdf"]:
        metric = hazard * survival  # f(t) = h(t) * S(t)
    else:
        raise ValueError("The measure '" + str(measure) + "' is not supported. "
                         "Please select one of the following: survival, density, risk, hazard, cumulative_hazard.")
    return metric



#####################################################################################################################
# Estimating functions for measurement error
#####################################################################################################################

import numpy as np

from delicatessen.utilities import inverse_logit
from delicatessen.estimating_equations.regression import ee_regression
from delicatessen.estimating_equations.processing import generate_weights


def ee_rogan_gladen(theta, y, y_star, r, weights=None):
    r"""Estimating equation for the Rogan-Gladen correction for mismeasured *binary* outcomes. This estimator uses
    external data to estimate the sensitivity and specificity, and then uses those external estimates to correct the
    estimated proportion. The general form of the estimating equations are

    .. math::

        \sum_{i=1}^n
        \begin{bmatrix}
            \mu \times \left\{ \alpha + \beta - 1 \right\} - \left\{ \mu^* + \beta - 1 \right\} \\
            R_i (Y_i^* - \mu^*) \\
            (1-R_i) Y_i \left\{ Y^*_i - \beta \right\} \\
            (1-R_i) (1-Y_i) \left\{ (1 - Y^*_i) - \alpha \right\} \\
        \end{bmatrix}
        = 0

    where :math:`Y` is the true value of the outcome, :math:`Y^*` is the mismeasured value of the outcome, :math:`R` is
    the indicator for the main study data, :math:`\mu` is the corrected mean, :math:`\mu^*` is the mismeasured mean in
    the main study data, :math:`\beta` is the sensitivity, and :math:`\alpha` is the specificity. The first
    estimating equation is the corrected proportion, the second is the naive proportion, the third is for sensitivity,
    and the fourth for specificity.

    Here, ``theta`` is a 1-by-4 array.

    Note
    ----
    The Rogan-Gladen estimator may provide corrected proportions outside of :math:`[0,1]` when
    :math:`\alpha + \beta \le 1`.

    Parameters
    ----------
    theta : ndarray, list, vector
        Theta consists of 4 values.
    y : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the gold-standard :math:`Y` measurements in the external
        sample. All values should be either 0 or 1, and be non-missing among those with :math:`R=0`.
    y_star : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the mismeasured :math:`Y` values. All values should be
        either 0 or 1, and be non-missing among all observations.
    r : ndarray, list, vector
        1-dimensional vector of `n` indicators regarding whether an observation was part of the external validation
        data. Indicator should designate if observations are the main data.
    weights : ndarray, list, vector, None, optional
        1-dimensional vector of `n` weights. Default is ``None``, which assigns a weight of 1 to all observations.

    Returns
    -------
    array :
        Returns a 4-by-`n` NumPy array evaluated for the input ``theta``

    Examples
    --------
    Construction of a estimating equation(s) with ``ee_rogan_gladen`` should be done similar to the following

    >>> import numpy as np
    >>> import pandas as pd
    >>> from scipy.stats import logistic
    >>> from delicatessen import MEstimator
    >>> from delicatessen.estimating_equations import ee_rogan_gladen

    Replicating the published example from Cole et al. (2023).

    >>> d = pd.DataFrame()
    >>> d['Y_star'] = [0, 1] + [0, 1, 0, 1]
    >>> d['Y'] = [np.nan, np.nan] + [0, 0, 1, 1]
    >>> d['S'] = [1, 1] + [0, 0, 0, 0]
    >>> d['n'] = [270, 680] + [71, 18, 38, 203]
    >>> d = pd.DataFrame(np.repeat(d.values, d['n'], axis=0), columns=d.columns)

    Applying the Rogan-Gladen correction to this example

    >>> def psi(theta):
    >>>     return ee_rogan_gladen(theta=theta, y=d['Y'],
    >>>                            y_star=d['Y_star'], r=d['S'])

    Notice that ``y`` corresponds to the gold-standard outcomes (only available where R=0), ``y_star`` corresponds to
    the mismeasured covariate data (available for R=1 and R=0), and ``r`` corresponds to the indicator for the main
    data source. Now we can call the M-Estimator.

    >>> estr = MEstimator(psi, init=[0.5, 0.5, .75, .75])
    >>> estr.estimate(solver='lm')

    Inspecting the parameter estimates, variance, and 95% confidence intervals

    >>> estr.theta
    >>> estr.variance
    >>> estr.confidence_intervals()

    The corrected proportion is

    >>> estr.theta[0]

    Inverse probability weights can be used through the ``weights`` argument. See the applied examples for a
    demonstration.

    References
    ----------
    Cole SR, Edwards JK, Breskin A, Rosin S, Zivich PN, Shook-Sa BE, & Hudgens MG. (2023). Illustration of 2 Fusion
    Designs and Estimators. *American Journal of Epidemiology*, 192(3), 467-474.

    Rogan WJ & Gladen B. (1978). Estimating prevalence from the results of a screening test.
    *American Journal of Epidemiology*, 107(1), 71-76.

    Ross RK, Zivich PN, Stringer JSA, & Cole SR. (2024). M-estimation for common epidemiological measures: introduction
    and applied examples. *International Journal of Epidemiology*, 53(2), dyae030.
    """
    # Processing inputs
    y = np.asarray(y)                           # Convert to NumPy array
    y_star = np.asarray(y_star)                 # Convert to NumPy array
    r = np.asarray(r)                           # Convert to NumPy array
    y = np.where(r == 1, -999, y)               # Removing NaN (or any other indicators) for Y in main
    w = generate_weights(weights=weights,       # Formatting weights
                         n_obs=y.shape[0])      # ... and processing if None
    mu = theta[0]                               # Parameter for corrected proportion
    mu_star = theta[1]                          # Means (corrected and naive)
    sens = theta[2]                             # Parameters for the sensitivity model
    spec = theta[3]                             # Parameters for the specificity model

    # Estimating equation for naive (mismeasured) mean
    ee_naive_mean = r * w * (y_star - mu_star)  # Naive mean among obs

    # Estimating equation for sensitivity
    ee_sens = (y_star - sens) * (1-r) * y * w

    # Estimating equation for specificity
    ee_spec = (1 - y_star - spec) * (1-r) * (1-y) * w

    # Estimating equation for the Rogan-Gladen correction
    ee_corr_mean = mu*(sens + spec - 1) - (mu_star + spec - 1) * np.ones(y.shape[0])

    # Returning stacked estimating equations
    return np.vstack([ee_corr_mean,             # Corrected mean
                      ee_naive_mean,            # Naive mean
                      ee_sens,                  # Sensitivity model parameters
                      ee_spec])                 # Specificity model parameters


def ee_rogan_gladen_extended(theta, y, y_star, r, X, weights=None):
    r"""Estimating equation for the extended Rogan-Gladen correction for mismeasured *binary* outcomes. This estimator
    uses external data to estimate the sensitivity and specificity conditional on covariates, and then uses those
    external estimates to correct the estimated proportion. The general form of the estimating equations are

    .. math::

        \sum_{i=1}^n
        \begin{bmatrix}
            R_i \times \left\{ \frac{Y^* + m(X_i; \beta) - 1}{m(X_i; \alpha) + m(X_i; \beta) - 1}  - \mu \right\} \\
            (1-R_i) Y_i \left\{ Y^*_i - m(X_i; \beta) \right\} X_i^T \\
            (1-R_i) (1 - Y_i) \left\{ (1 - Y^*_i) - m(X_i; \beta) \right\} X_i^T \\
        \end{bmatrix}
        = 0

    where :math:`Y` is the true value of the outcome, :math:`Y^*` is the mismeasured value of the outcome. The first
    estimating equation is the corrected proportion, the second is for sensitivity, and the third for specificity.

    If :math:`X` is of dimension :math:`p`, then ``theta`` is a 1-by-(1+2`p`) array. Note that the design matrix is
    shared across the sensitivity and specificity models.

    Note
    ----
    The Rogan-Gladen estimator may provide corrected proportions outside of :math:`[0,1]` when
    :math:`\alpha + \beta \le 1`, or the addition of sensitivity and specificity is less than or equal to one.

    Parameters
    ----------
    theta : ndarray, list, vector
        Theta consists of 1+2`p` values.
    y : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the gold-standard :math:`Y` measurements in the external
        sample. All values should be either 0 or 1, and be non-missing among those with :math:`R=0`.
    y_star : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the mismeasured :math:`Y` values. All values should be
        either 0 or 1, and be non-missing among all observations.
    r : ndarray, list, vector
        1-dimensional vector of `n` indicators regarding whether an observation was part of the external validation
        data. Indicator should designate if observations are the main data.
    X : ndarray, list, vector
        2-dimensional vector of a design matrix for the sensitivity and specificity models.
    weights : ndarray, list, vector, None, optional
        1-dimensional vector of `n` weights. Default is ``None``, which assigns a weight of 1 to all observations.

    Returns
    -------
    array :
        Returns a (1+2`p`)-by-`n` NumPy array evaluated for the input ``theta``

    Examples
    --------
    Construction of a estimating equation(s) with ``ee_rogan_gladen_extended`` should be done similar to the following

    >>> import numpy as np
    >>> import pandas as pd
    >>> from scipy.stats import logistic
    >>> from delicatessen import MEstimator
    >>> from delicatessen.estimating_equations import ee_rogan_gladen_extended

    Replicating the example from Cole et al. (2023).

    >>> d = pd.DataFrame()
    >>> d['Y_star'] = [0, 1] + [0, 1, 0, 1]
    >>> d['Y'] = [np.nan, np.nan] + [0, 0, 1, 1]
    >>> d['S'] = [1, 1] + [0, 0, 0, 0]
    >>> d['n'] = [270, 680] + [71, 18, 38, 203]
    >>> d = pd.DataFrame(np.repeat(d.values, d['n'], axis=0), columns=d.columns)
    >>> d['C'] = 1

    Applying the Rogan-Gladen correction to this example

    >>> def psi(theta):
    >>>     return ee_rogan_gladen_extended(theta=theta, y=d['Y'],
    >>>                                     y_star=d['Y_star'],
    >>>                                     X=d[['C', ]], r=d['S'])

    Notice that ``y`` corresponds to the gold-standard outcomes (only available where R=0), ``y_star`` corresponds to
    the mismeasured covariate data (available for R=1 and R=0), and ``r`` corresponds to the indicator for the main
    data source. Now we can call the M-Estimator.

    >>> estr = MEstimator(psi, init=[0.5, 1., 1.])
    >>> estr.estimate(solver='lm')

    Inspecting the parameter estimates, variance, and 95% confidence intervals

    >>> estr.theta
    >>> estr.variance
    >>> estr.confidence_intervals()

    Note
    ----
    The sensitivity and specificity in ``ee_rogan_gladen_extended`` correspond to the logit transformations, unlike
    ``ee_rogan_gladen`` which returns the sensitivity and specificity directly.


    The corrected proportion is

    >>> estr.theta[0]

    Inverse probability weights can be used through the ``weights`` argument. See the applied examples for a
    demonstration.

    References
    ----------
    Cole SR, Edwards JK, Breskin A, Rosin S, Zivich PN, Shook-Sa BE, & Hudgens MG. (2023). Illustration of 2 Fusion
    Designs and Estimators. *American Journal of Epidemiology*, 192(3), 467-474.

    Rogan WJ & Gladen B. (1978). Estimating prevalence from the results of a screening test.
    *American Journal of Epidemiology*, 107(1), 71-76.

    Ross RK, Cole SR, Edwards JK, Zivich PN, Westreich D, Daniels JL, Price JT & Stringer JSA. (2024). Leveraging
    External Validation Data: The Challenges of Transporting Measurement Error Parameters. *Epidemiology*,
    35(2), 196-207.
    """
    # Processing inputs
    y = np.asarray(y)                           # Convert to NumPy array
    y_star = np.asarray(y_star)                 # Convert to NumPy array
    r = np.asarray(r)                           # Convert to NumPy array
    X = np.asarray(X)                           # Convert to NumPy array
    if weights is None:                         # Handle weights argument
        weights = 1                             # ... set all weight as 1
    else:                                       # Otherwise
        weights = np.asarray(weights)           # ... convert to NumPy array

    # Preparing data for estimating equation operations
    nXp = X.shape[1] + 1                        # Index start for the NumPy matrices
    y = np.where(r == 1, -999, y)               # Removing NaN (or any other indicators) for Y in main
    mu = theta[0]                               # Parameter of interest
    sens = theta[1:nXp]                         # Parameters for sensitivity model
    spec = theta[nXp:]                          # Parameters for specificity model

    # Nuisance models for sensitivity
    ee_sens = ee_regression(theta=sens, y=y_star, X=X,
                            model='logistic', weights=weights) * (1-r) * y
    sens_i = inverse_logit(np.dot(X, sens))     # Predicted sensitivity for each unit

    # Nuisance models for specificity
    ee_spec = ee_regression(theta=spec, y=1-y_star, X=X,
                            model='logistic', weights=weights) * (1-r) * (1-y)
    spec_i = inverse_logit(np.dot(X, spec))     # Predicted specificity for each unit

    # Estimating equation for the individual-level version of the Rogan-Gladen correction
    rg_equation = (y_star + spec_i - 1) / (sens_i + spec_i - 1)
    ee_corr_mean = r * (rg_equation - mu) * weights

    # Returning the stacked estimating equations
    return np.vstack([ee_corr_mean, ee_sens, ee_spec])


def ee_regression_calibration(theta, beta, a, a_star, r, X=None, weights=None):
    r"""Estimating equation for regression calibration with external data for a mismeasured *binary* action. Regression
    calibration is a simple to implement method to correct for measurement error.

    The general form of the estimating equations are

    .. math::

        \sum_{i=1}^n
        \begin{bmatrix}
            (\beta^* / \gamma_0) - \beta  \\
            (1-R_i) \left\{ A_i - \gamma_0 A_i^* + \gamma^T X_i \right\} X_i^T
        \end{bmatrix}
        = 0

    where :math:`A` is the gold-standard measurement of the action, :math:`A^*` is the mismeasured version of the binary
    action, :math:`X` is some additional covariates (including at least an intercept), and :math:`R` indicates whether
    someone was in the validation set (:math:`R=0` if in the validation set).

    The first estimating equation is for the corrected coefficient for :math:`A` on :math:`Y`. This is done by scaling
    the coefficient for :math:`A^*` on :math:`Y` (which comes from a model external to ``ee_regression_calibration``)
    by the predictiveness in terms of probability of :math:`A^*` for :math:`A`, :math:`\gamma_0`. The second estimating
    equation is used to estimate :math:`\gamma` using a linear probability model. Here, :math:`\gamma` are the
    parameters of the calibration model. Notice that only the validation set contributes to the calibration model.

    Note
    ----
    For the second place in ``theta``, (i.e., ``theta[1]``), a starting value between between 0.5 and 1 is recommended.


    One caution for application of regression calibration is that it is only valid for non-differential measurement
    error. In cases of differential measurement error, methods like Multiple Imputation for Measurement Error (MIME)
    should be considered instead (Cole et al., 2006).

    If ``X=None`` then ``theta`` is a 1-by-3 array. Otherwise, ``theta`` is a 1-by-(2+`p`) array, where `p` is the is
    the dimension of :math:`X`.

    Parameters
    ----------
    theta : ndarray, list, vector
        Theta consists of 1+2`p` values.
    beta : float, int, ndarray
        Coefficient to correct from a model fit outside of ``ee_regression_calibration``. This coefficient should be
        for the main effect of ``a_star`` on the outcome. Notice that regression calibration only requires the
        coefficient to apply the correction (i.e., ``y`` is not needed for this estimating equation).
    a : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the gold-standard :math:`A` measurements in the external
        sample. All values should be either 0 or 1, and be non-missing among those with :math:`R=0`.
    a_star : ndarray, list, vector
        1-dimensional vector of `n` observed values. These are the mis-measured :math:`A` in the external and internal
        sample. All values should be either 0 or 1, and must be non-missing among those with :math:`R=0`.
    r : ndarray, list, vector
        1-dimensional vector of `n` indicators regarding whether an observation was part of the external validation
        data. Indicator should designate if observations are the main data.
    X : ndarray, list, vector, None, optional
        2-dimensional vector of a design matrix for calibration model. Notice that this design matrix should not include
        ``a``. Behind the scenes, ``a`` is added to this design matrix to make it easier to process the coefficients for
        the regression calibration step. Default is ``None``, which automatically generates an intercept, so the
        calibration model ``a ~ a_star + 1`` is fit by default.
    weights : ndarray, list, vector, None, optional
        1-dimensional vector of `n` weights. Default is ``None``, which assigns a weight of 1 to all observations. Note
        that weights are only used in the calibration model fitting.

    Returns
    -------
    array :
        Returns a 3-by-`n` or (2+`p`)-by-`n` NumPy array evaluated for the input ``theta``

    Examples
    --------
    Construction of a estimating equation(s) with ``ee_regression_calibration`` should be done similar to the following

    >>> import numpy as np
    >>> import pandas as pd
    >>> from scipy.stats import logistic
    >>> from delicatessen import MEstimator
    >>> from delicatessen.estimating_equations import ee_regression
    >>> from delicatessen.estimating_equations import ee_regression_calibration

    Creating some example data for illustration

    >>> d = pd.DataFrame()
    >>> d['A_star'] = [0, 1, 0, 1] + [1, 0, 1, 0]
    >>> d['A'] = [np.nan, np.nan, np.nan, np.nan] + [1, 1, 0, 0]
    >>> d['Y'] = [0, 0, 1, 1] + [0, 1] + [np.nan, np.nan]
    >>> d['n'] = [266, 67, 400, 267] + [90, 10, 20, 80]
    >>> d['S'] = [1, 1, 1, 1] + [0, 0, 0, 0]
    >>> d = pd.DataFrame(np.repeat(d.values, d['n'], axis=0), columns=d.columns)
    >>> d['C'] = 1
    >>> d1 = d.loc[d['S'] == 1].copy()

    Suppose we are interested in the odds ratio for A on Y. In the main study, we only have a mismeasured version of A,
    indicated by ``A_star``. As a starting point, we might examine ``A_star`` on Y, which can be done using a logistic
    regression model with delicatessen

    >>> def psi(theta):
    >>>     return ee_regression(theta=theta, y=d1['Y'],
    >>>                          X=d1[['C', 'A_star']],
    >>>                          model='logistic')

    >>> estr = MEstimator(psi, init=[0, 0])
    >>> estr.estimate()
    >>> np.exp(estr.theta[1])  # Odds Ratio of interest

    While we obtain a result, the corresponding odds ratio may be biased due to measurement error of A. Under the
    assumption of non-differential measurement error, we can use regression calibration (and our external data) to
    correct for the measurement error.

    To implement regression calibration, we stack our previous estimating equations with the regression calibration
    estimating equations. Below is code that illustrates this process

    >>> def psi(theta):
    >>>     theta_calib = theta[:3]  # Calibration model parameters
    >>>     theta_main = theta[3:]   # Naive model parameters
    >>>     y_no_nan = np.where(d['Y'].isna(), -999, d['Y'])
    >>>     r = np.asarray(d['S'])
    >>>     # Naive regression model
    >>>     ee_logit_star = ee_regression(theta=theta_main, y=y_no_nan,
    >>>                                   X=d[['C', 'A_star']],
    >>>                                   model='logistic')
    >>>     ee_logit_star = ee_logit_star * r  # Only main-study contributes
    >>>     # Regression calibration
    >>>     ee_calib = ee_regression_calibration(theta=theta_calib,  # Calibration parameters
    >>>                                          beta=theta_main[1], # Coefficient to correct
    >>>                                          a=d['A'],           # True A
    >>>                                          a_star=d['A_star'], # Mismeasured A
    >>>                                          r=r)                # Validation set indicator
    >>>     return np.vstack([ee_calib, ee_logit_star])

    >>> estr = MEstimator(psi, init=[0, 0.75, 0, 0, 0])
    >>> estr.estimate()

    Note that 5 initial starting values are provided, 3 for the regression calibration equations and 2 for the main
    model. Further, note that the second parameter is non-zero. Providing a starting value of zero will result in a
    division by zero error. It is recommended to have the starting value between 0.5 and 1. Finally, note that the
    naive log-odds ratio is provided to ``ee_regression_calibration`` via the ``beta`` argument. This argument tells
    ``ee_regression_calibration`` which coefficient needs to be corrected.

    Inspecting the parameter estimates

    >>> estr.theta[0]                    # Corrected log-odds ratio
    >>> estr.theta[1]                    # Calibration factor for the coefficient
    >>> estr.theta[-1]                   # Uncorrected log-odds ratio (for this setup)
    >>> estr.theta[-1] / estr.theta[1]   # Equivalent to estr.theta[0]

    While one could calculate the corrected odds ratio by-hand, the sandwich variance estimator provides a consistent
    estimate of the variance in this context. The provided estimating equation  method can be easily extended to
    account for measurement error of conditional odds ratios, parameters of generlized linear models, or marginal
    structural models.

    Weighted calibration models can be estimated by specifying the optional ``weights`` argument (weights are not used
    to calibrate the coefficient).

    References
    ----------
    Boe LA, Lumley T, & Shaw PA. (2024). Practical Considerations for Sandwich Variance Estimation in 2-Stage
    Regression Settings. *American Journal of Epidemiology*, 193(5), 798-810.

    Cole SR, Chu H, & Greenland S. (2006). Multiple-imputation for measurement-error correction.
    *International Journal of Epidemiology*, 35(4), 1074-1081.

    Cole SR, Jacobson LP, Tien PC, Kingsley L, Chmiel JS, & Anastos K. (2010). Using marginal structural
    measurement-error models to estimate the long-term effect of antiretroviral therapy on incident AIDS or death.
    *American Journal of Epidemiology*, 171(1), 113-122.

    Rosner B, Willett WC, & Spiegelman D. (1989). Correction of logistic regression relative risk estimates and
    confidence intervals for systematic within‐person measurement error. *Statistics in Medicine*, 8(9), 1051-1069.
    """
    # Processing inputs
    a = np.asarray(a)                           # Convert to NumPy array
    a_star = np.asarray(a_star)                 # Convert to NumPy array
    r = np.asarray(r)                           # Convert to NumPy array
    if X is None:                               # Intercept-only model
        X = np.ones(a.shape)[:, None]           # ... intercept-only
    else:                                       # User-specified design matrix
        X = np.asarray(X)                       # ... convert to NumPy array
    if weights is None:                         # Handle weights argument
        weights = 1                             # ... set all weight as 1
    else:                                       # Otherwise
        weights = np.asarray(weights)           # ... convert to NumPy array

    # Preparing data for estimating equation operations
    a = np.where(r == 1, -999, a)               # Removing NaN (or any other indicators) for Y in main
    beta_corrected = theta[0]                   # First parameter in vector will be corrected coefficient
    gamma = theta[1:]                           # All other parameters are for the regression calibration model

    # Calibration Model (via a linear probability model)
    ee_calib = ee_regression(theta=gamma,                              # Regression model for the calibration step
                             y=a, X=np.hstack([a_star[:, None], X]),   # ... Design matrices built as I expect them
                             model='linear',                           # ... linear regression
                             weights=weights)                          # ... with provided weights
    ee_calib = ee_calib * (1-r)                                        # Only have the external observations contribute

    # Corrected coefficient
    ee_corr_beta = np.ones(a.shape) * (beta / gamma[0] - beta_corrected)  # Transformation of coefficients for RC

    # Returning the stacked estimating equations
    return np.vstack([ee_corr_beta, ee_calib])



import numpy
import numpy as np
import scipy as sp
from scipy.stats import norm


def approx_differentiation(xk, f, epsilon=1e-9, method='capprox'):
    r"""Numerical approximation to compute the gradient. This function implements numerical approximation methods for
    derivatives generally (i.e., it provides the first-order forward, backward, and central difference approximations).

    Note
    ----
    This functionality is only intended for use behind the scenes in ``delicatessen``. Numerical approximation is
    implemented from scratch to offer backward and central difference approximations (SciPy's ``approx_fprime`` only
    offers the forward difference).


    The forward difference approximation is

    .. math::

        \frac{f(x + \epsilon) - f(x)}{\epsilon}

    the backward difference approximation is

    .. math::

        \frac{f(x) - f(x - \epsilon)}{\epsilon}

    and the central difference approximation is

    .. math::

        \frac{f(x + \epsilon) - f(x - \epsilon)}{2\epsilon}

    Here, the numerical approximation is implemented by generating matrices for output from a function evaluated under
    minor perturbations (determined by ``epsilon``) of each input argument. These matrices are then subtracted from
    each other and then scaled by ``epsilon``.

    Parameters
    ----------
    xk : ndarray, list, shape (n, )
        Point(s) or coordinate vector to evaluate the gradient at.
    f : callable
        Function of which to estimate the gradient of.
    epsilon : float, optional
        Increment to perturb the points by to compute the gradient. This should be a small value
    method : str, optional
        Approximation to use to compute the gradient. Default is `capprox` which uses the central difference method.
        One can also specify the forward difference (`fapprox`) or backward difference (`bapprox`) methods.

    Returns
    -------
    numpy.array :
        Corresponding array of the pairwise derivatives for all different input x values.

    Examples
    --------
    >>> import numpy as np
    >>> from delicatessen.derivative import approx_differentiation

    To illustrate use, we will compute the derivative of the following function

    .. math::

        f(x) = x^2 - x^1 + sin(x + \sqrt{x})

    >>> def f(x):
    >>>     return x**2 - x + np.sin(x + np.sqrt(x))

    If you work out the deriative by-hand, you will end up with the following

    .. math::

        2x - 1 + \left( \frac{1}{2 \sqrt{x}} + 1 \right) \cos(x + \sqrt{x})

    Instead, we can use the central difference approximation to evaluate the derivative at a specific point. Here, we
    will evaluate the derivative at :math:`x=1`

    >>> dy = approx_differentiation(xk=[1, ], f=f)

    which returns ``0.37578``, which is close to plugging in :math:`x=1` into the previous equation.

    The derivative of a function with multiple inputs and multiple outputs can also be evaluated. Consider the following
    example with three inputs and two outputs

    >>> def f(x):
    >>>     return [x[0]**2 - x[1], np.sin(np.sqrt(x[1]) + x[2]) + x[2]*(x[1]**2)]

    >>> approx_differentiation(xk=[0.7, 1.2, -0.9], f=f, method='fapprox')
    >>> approx_differentiation(xk=[0.7, 1.2, -0.9], f=f, method='bapprox')
    >>> approx_differentiation(xk=[0.7, 1.2, -0.9], f=f, method='capprox')

    which will return a 2-by-3 array of all the x-y pair derivatives at the given values. Here, the rows correspond to
    the output and the columns correspond to the inputs. The approximation methods are forward, backward, and central.
    """
    # Setup parameters for call
    xk = np.asarray(xk)                               # Convert inputs into NumPy array if not already
    xp = xk.shape[0]                                  # Get the number of parameters in the input
    shift = np.identity(n=xk.shape[0]) * epsilon      # Define the shift matrix for the partials

    def generate_matrix(x_shift, f):
        """Internal function to generate a matrix of the outputs under the parameter shifts, defined by x_shift"""
        shift_matrix = []                             # Storage for matrices
        for j in range(xp):                           # Looping over shift combinations
            shift_matrix.append(f(x_shift[j, :]))     # ... compute output at shifted values
        return np.asarray(shift_matrix)               # Return matrix under all shifts

    # Computing the gradient using the corresponding method
    if method == 'capprox':                           # Central difference
        lower = (xk - shift)                          # ... defining lower shift
        f0 = generate_matrix(x_shift=lower, f=f)      # ... output for lower shift
        upper = (xk + shift)                          # ... defining upper shift
        f1 = generate_matrix(x_shift=upper, f=f)      # ... output for upper shift
        deriv = (f1 - f0).T / (2*epsilon)             # ... central difference approximation
    elif method == 'bapprox':                         # Backward difference
        lower = (xk - shift)                          # ... defining lower shift
        f0 = generate_matrix(x_shift=lower, f=f)      # ... output for lower shift
        f_eval = f(xk)                                # ... upper is held fixed
        f1 = np.asarray([f_eval for i in range(xp)])  # ... stack upper into a matrix
        deriv = (f1 - f0).T / epsilon                 # ... backward difference approximation
    elif method == 'fapprox':                         # Forward difference
        f_eval = f(xk)                                # ... lower is held fixed
        f0 = np.asarray([f_eval for i in range(xp)])  # ... stack lower into a matrix
        upper = (xk + shift)                          # ... defining upper shift
        f1 = generate_matrix(x_shift=upper, f=f)      # ... output for upper shift
        deriv = (f1 - f0).T / epsilon                 # ... forward difference approximation
    else:                                             # Otherwise error
        raise ValueError("The chosen method `"
                         + str(method)
                         + "` chosen is not supported")

    # Processing the final return based on parameter shape
    if xp == 1:
        return np.asarray([deriv, ])
    else:
        return deriv


def auto_differentiation(xk, f):
    r"""Forward-mode automatic differentiation. Automatic differentiation offers a way to compute the exact derivative,
    rather than numerically approximated (i.e., the central difference method). Automatic differentiation iteratively
    applies the chain rule through recursive calls to evaluate the derivative.

    Note
    ----
    This functionality is only intended for use behind the scenes in ``delicatessen``. Automatic differentiation is
    implemented from scratch to avoid additional dependencies.


    This is accomplished by the ``PrimalTangentPairs`` class, which is a special data type in ``delicatessen`` that
    stores pairs of the original evaluation and the corresponding derivative for a variety of different mathematical
    operations. This is what allows for the exact derivative calculation. The ``auto_differentiation`` function is
    a wrapper to access and use this class object as it is intended for derivative computations.

    Parameters
    ----------
    xk : ndarray, list, shape (n, )
        Point(s) or coordinate vector to evaluate the gradient at.
    f : callable
        Function of which to estimate the gradient of.

    Returns
    -------
    numpy.array :
        Corresponding array of the pairwise derivatives for all different input x values.

    Examples
    --------
    Loading necessary functions

    >>> import numpy as np
    >>> from delicatessen.derivative import auto_differentiation

    To illustrate use, we will compute the derivative of the following function

    .. math::

        f(x) = x^2 - x^1 + sin(x + \sqrt{x})

    >>> def f(x):
    >>>     return x**2 - x + np.sin(x + np.sqrt(x))

    If you work out the deriative by-hand, you will end up with the following

    .. math::

        2x - 1 + \left( \frac{1}{2 \sqrt{x}} + 1 \right) \cos(x + \sqrt{x})

    Instead, we can use automatic differentiation to evaluate the derivative at a specific point. Here, we will
    evaluate the derivative at :math:`x=1`

    >>> dy = auto_differentiation(xk=[1, ], f=f)

    which returns ``0.3757795``. This is the same as if you plugged in :math:`x=1` into the previous equation.

    Note
    ----
    If a derivative is not defined, then the function will return a ``NaN``.


    The derivative of a function with multiple inputs and multiple outputs can also be evaluated. Consider the following
    example with three inputs and two outputs

    >>> def f(x):
    >>>     return [x[0]**2 - x[1], np.sin(np.sqrt(x[1]) + x[2]) + x[2]*(x[1]**2)]

    >>> dy = auto_differentiation(xk=[0.7, 1.2, -0.9], f=f)

    which will return a 2-by-3 array of all the x-y pair derivatives at the given values. Here, the rows correspond to
    the output and the columns correspond to the inputs.

    References
    ----------
    Baydin AG, Pearlmutter BA, Radul AA, & Siskind JM. (2018). Automatic differentiation in machine learning: a survey.
    *Journal of Marchine Learning Research*, 18, 1-43.

    Rall LB & Corliss GF. (1996). An introduction to automatic differentiation. Computational Differentiation:
    Techniques, Applications, and Tools, 89, 1-18.
    """
    # Meta-information about function and inputs
    xshape = len(xk)                                                   # The number of inputs into the function

    # Set up Pair objects for evaluating gradient of function
    pairs_for_gradient = []                                            # Storage for the pairs to provide function
    for i in range(xshape):                                            # For each of the inputs
        partial = np.zeros_like(xk)                                    # ... generate array of that size of all zeroes
        partial[i] = 1                                                 # ... replace 0 with 1 for current index
        pairs_for_gradient.append(PrimalTangentPairs(xk[i], partial))  # ... then store as a primal,tangent pair
    x_to_eval = np.asarray(pairs_for_gradient)                         # Convert from list to NumPy array

    # Internal function to handle some specific exceptions when computing derivatives
    def f_deny_bool(function, x):
        """This internal function re-evaluates the input function with an additional operator. Namely zero is added
        to the function. This causes no difference in the primal or tangent pair. The purpose of this function is to
        work around a specific behavior of ``np.where(...)``.

        The function ``np.where`` can a weird call order to the standard Python operators. It calls ``__bool__`` as the
        final operation. This means that ``PrimalTangentPairs`` can only return a ``bool`` for ``np.where`` calls.
        Consider the following example

        .. code::

            np.where(x[0] >= 1, x[1]**2, 0)


        If ``x[0] >= 1`` then we should find the derivative of ``x[1]**2``. For ``np.where``, ``x[1]**2`` is evaluated
        before ``x[0] >=1``. As a result of the boolean piece being the last part evaluated, resulting in an empty
        array.

        By adding zero, we ensure that the final operation applied is an addition, and thus an float is returned
        after the boolean is evaluated. Essentially, we deny the call to ``__bool__`` being the last operation.
        """
        evalued = function(x)              # Short-name for evaluated function

        # Handling whether a single input (non-tuple like SciPy) is provided
        if isinstance(evalued, PrimalTangentPairs):     # If not a tuple, then gives back the Pair object
            eval_no_bool_end = evalued + 0              # ... then just add zero in this case
        else:                                           # Otherwise,
            eval_no_bool_end = []                       # ... empty list for storage
            for e in evalued:                           # ... for each evaluation in the function
                eval_no_bool_end.append(e+0)            # ... adding zero in a loop

        return eval_no_bool_end            # Return the evaluation with the final addition

    # Evaluating the function for the primal,tangent pairs
    evaluated_pair = f_deny_bool(function=f,                           # Evaluate the primal,tangent pair with function
                                 x=x_to_eval)                          # ... at the given values

    # Processing function output into gradient value or matrix
    evaluated_gradient = []                                            # List storage for the computed gradient
    if isinstance(evaluated_pair, PrimalTangentPairs):                 # Handle case where only single input
        evaluated_gradient.append([evaluated_pair.tangent, ])          # ... evaluate tangent and put list in list
    else:                                                              # In all other cases
        for pair in evaluated_pair:                                    # ... for each evaluated primal, tangent pair
            if isinstance(pair, PrimalTangentPairs):                   # ... if that pair is the key type
                evaluated_gradient.append(pair.tangent)                # ... then give back the derivative
            else:                                                      # ... otherwise row has no xk operations
                empty_array = np.array([0 for j in range(xshape)])     # ... create array of all zeroes
                evaluated_gradient.append(empty_array)                 # ... so derivative is always zero

    # Return evaluated gradient as a NumPy array
    if len(evaluated_gradient) == 1:                                   # If only consists of 1 item
        return np.asarray(evaluated_gradient[0])                       # ... return that item as NumPy array
    else:                                                              # Otherwise
        return np.asarray(evaluated_gradient)                          # ... return is as an array


class PrimalTangentPairs:
    """Unique class object for automatic differentiation. This class divides the inputs into 'primal' and 'tangent'
    pairs. The derivative is computed via operations on the pairs, which are then recurvsively called by this class.
    This process allows use to successively apply the chain rule.

    Note
    ----
    This class only handles how the data is setup (i.e., it does not automatically compute a derivative. The function
    for the derivative computation then takes this as input.


    This data class is only meant to interact with ``auto_differentiation`` and should not be used elsewhere.

    Parameters
    ----------
    primal :
        The x values for which the derivative computation is desired. Must be the same length as ``tangent``.
    tangent :
        Indicator for the location at which the derivatives is desired. Must be the same length as ``primal``.
    """
    # Some internal notes on handling ndarray's in this class (because there are some tricky issues). First, we do
    #   not want to override the priority of the NumPy operators (in fact we rely on it to take precedence). This is
    #   extremely important when trying to mix operators together that start with a numpy.ndarray, as we want
    #   everything to go to element-wise operations. So, we must not do something like the following
    # __array_priority__ = 2
    # Instead, we have modified all the operations that include `other` to check if the input is a np.ndarray. If it is
    #   (i.e., the np.ndarray is the second object in x [] y) then we manually reverse the operations so that the
    #   np.ndarray leads (i.e., y [] x). This gets NumPy to decompose everything correctly to element-wise operations.
    #   The downside is that this introduces floating point errors for powers, when it is being raised to a np.ndarray
    #   power and everything is integers. This floating point error was determined to be acceptable to get everything
    #   working for the automatic differentation.

    def __init__(self, primal, tangent):
        # Processing of the inputs into the class, both initial and recursive calls
        if isinstance(primal, PrimalTangentPairs):    # If given a PrimalTangentPair input
            if isinstance(primal.primal, np.ndarray):
                raise ValueError("... order of operations issue... I am getting an array an input")
            self.primal = primal.primal               # ... extract the primal element from input
        else:                                         # Else
            if isinstance(primal, np.ndarray):
                raise ValueError("... order of operations issue... I am getting an array an input")
            self.primal = primal                      # ... directly save as new primal
        self.tangent = tangent                        # Store the tangent

    # Basic operators for the class object

    def __str__(self):
        # Conversion to string just to in case it gets called somehow
        return f"PrimalTangentPairs({self.primal}, {self.tangent})"

    def __bool__(self):
        # To get np.where working properly, I need to have the internal boolean function for this class return
        #   only the primal part. This is a bool object type, which is what is expected and has np.where operate as
        #   expected. This only seems to be called for np.where and not the other operators (they work directly).
        return bool(self.primal)

    def transpose(self):
        # Transpose operator is special
        storage = []  # Create empty storage for the output

        if self.tangent.ndim == 1:  # When tangent is of 1 dimension
            for p in self.primal:  # ... then only the primal's are a stack
                storage.append(PrimalTangentPairs(p,  # ... so we reshape the primal stack
                                                  self.tangent))  # ... and keep all the corresponding tangent's
        else:  # Otherwise, tangent is also 2 dimensions
            for p, t in zip(self.primal, self.tangent):  # ... so both the primal and tangents
                storage.append(PrimalTangentPairs(p, t))  # ... need to be re-paired

        # Return the transposed array of the PrimalTangentPairs
        return np.array(storage)

    @property
    def T(self):
        # Giving the transpose() property the .T short-cut, like NumPy has (to be compatible with NumPy code)
        return self.transpose()

    # Equality operators for the class

    def __le__(self, other):
        # Less than or equal to operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal <= other.primal)             # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal <= other)                    # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    def __lt__(self, other):
        # Less than operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal < other.primal)              # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal < other)                     # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    def __ge__(self, other):
        # Greater than or equal to operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal >= other.primal)             # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal >= other)                    # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    def __gt__(self, other):
        # Greater than operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal > other.primal)              # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal > other)                     # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    def __eq__(self, other):
        # Equality operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal == other.primal)             # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal == other)                    # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    def __ne__(self, other):
        # Inequality operator
        if isinstance(other, PrimalTangentPairs):                  # If the input is a PrimalTangentPairs
            comparison = (self.primal != other.primal)             # ... compare the primals
        else:                                                      # Otherwise
            comparison = (self.primal != other)                    # ... directly compare primal with other
        return PrimalTangentPairs(comparison, 0)                   # Return evaluated equality

    # Elementary mathematical operations

    def __neg__(self):
        # Negation operator
        return PrimalTangentPairs(-self.primal,                    # Negation is simply negation
                                  -self.tangent)                   # ... and same for derivative

    def __add__(self, other):
        # Addition
        if isinstance(other, PrimalTangentPairs):                     # When adding PrimalTangentPairs
            return PrimalTangentPairs(self.primal + other.primal,     # ... add primals together
                                      self.tangent + other.tangent)   # ... and add tangents together
        elif isinstance(other, np.ndarray):
            return other + self
        else:                                                         # Otherwise
            return PrimalTangentPairs(self.primal + other,            # ... add the primal and other
                                      self.tangent)                   # ... tangent does not change by constants

    def __radd__(self, other):
        # Addition, reversed
        return self.__add__(other)                                    # Reversed addition (order matters for process)

    def __sub__(self, other):
        # Subtraction
        if isinstance(other, PrimalTangentPairs):                     # When subtracting PrimalTangentPairs
            return PrimalTangentPairs(self.primal - other.primal,     # ... minus primals together
                                      self.tangent - other.tangent)   # ... and minus tangets together
        elif isinstance(other, np.ndarray):
            return -1*other + self
        else:                                                         # Otherwise
            return PrimalTangentPairs(self.primal - other,            # ... minus primal and other
                                      self.tangent)                   # ... tangent does not change by constants

    def __rsub__(self, other):
        # Subtraction, reversed
        return PrimalTangentPairs(other - self.primal,                # Subtracting in reverse order (lead constant)
                                  -1 * self.tangent)                  # ... simply multiply tangent by -1

    def __mul__(self, other):
        # Multiplication
        if isinstance(other, PrimalTangentPairs):                     # When multiplying PrimalTangentPairs
            return PrimalTangentPairs(self.primal * other.primal,     # ... multiply primals
                                      self.tangent*other.primal + self.primal*other.tangent)
        elif isinstance(other, np.ndarray):                           # When comparing with np.ndarray
            return other * self                                       # ... reverse the order to make operations work
        else:                                                         # Otherwise
            return PrimalTangentPairs(self.primal * other,            # ... multiply primal and constant
                                      self.tangent * other)           # ... multiply primal and constant

    def __rmul__(self, other):
        # Multiplication, reversed
        return PrimalTangentPairs(self.primal * other,                # Reversed order can evaluate the same
                                  self.tangent * other)               # ... for primal and tangent pairs

    def __truediv__(self, other):
        # Division
        if isinstance(other, PrimalTangentPairs):                     # When dividing PrimalTangentPairs
            return PrimalTangentPairs(self.primal / other.primal,     # ... divide primals
                                      (self.tangent * other.primal - self.primal * other.tangent) / (other.primal ** 2))
        elif isinstance(other, np.ndarray):                           # When comparing with np.ndarray
            return (1/other) * self                                   # ... reverse the order to make operations work
        else:                                                         # Otherwise
            return PrimalTangentPairs(self.primal / other,            # ... divide primal and constant
                                      self.tangent / other)           # ... divide tangent and constant

    def __rtruediv__(self, other):
        # Division, reversed
        return PrimalTangentPairs(other / self.primal,                # Reversed form of division
                                  (0 * self.primal - self.tangent * other) / (self.primal ** 2))

    def __pow__(self, other):
        # Power
        if isinstance(other, PrimalTangentPairs):                                    # For PrimalTangentPairs to powers
            # I'm not sure about allowing this option. So it is disallowed for the time being
            # if self.primal == 0:
            #     return PrimalTangentPairs(0, self.tangent*0)
            fgx = other.primal * self.primal ** (other.primal - 1) * self.tangent    # ... initial piece of rule
            gx = self.primal ** other.primal * np.log(self.primal) * other.tangent   # ... secondary piece of rule
            return PrimalTangentPairs(self.primal ** other.primal,                   # ... raise primal to primal
                                      fgx + gx)                                      # ... apply the power rule
        elif isinstance(other, np.ndarray):
            if self.primal >= 0:
                return np.exp(other * np.log(self))
            else:
                return np.where(other % 2 == 1,
                                -np.exp(other * np.log(-self)),
                                np.exp(other * np.log(-self)))
        else:
            return PrimalTangentPairs(self.primal ** other,                          # Otherwise compute directly
                                      other * self.primal ** (other - 1) * self.tangent)

    def __rpow__(self, other):
        # Power, reversed
        return PrimalTangentPairs(other ** self.primal,                                 # Constant to primal
                                  self.tangent * other ** self.primal * np.log(other))  # ... then power rule

    # Operators that sometimes have undefined derivatives

    def __abs__(self):
        # Absolute value
        if self.primal > 0:                                                            # Absolute value for positive
            return PrimalTangentPairs(1 * self.primal, self.tangent)                   # ... primal, tangent
        elif self.primal < 0:                                                          # Absolute value for negative
            return PrimalTangentPairs(-1 * self.primal, -1 * self.tangent)             # ... -primal, -tangent
        else:                                                                          # Absolute value at zero
            return PrimalTangentPairs(0 * self.primal, np.nan * self.tangent)          # ... zero, undefined

    def __floor__(self):
        # Floor
        if self.primal % 1 == 0:                                                       # Floor with integers
            return PrimalTangentPairs(np.floor(self.primal), np.nan * self.tangent)    # ... floor, undefined
        else:                                                                          # Floor with remainder
            return PrimalTangentPairs(np.floor(self.primal), 0 * self.tangent)         # ... floor, 0

    def __ceil__(self):
        # Ceiling
        if self.primal % 1 == 0:                                                       # Ceil with integers
            return PrimalTangentPairs(np.ceil(self.primal), np.nan * self.tangent)     # ... ceil, undefined
        else:                                                                          # Ceil with remainder
            return PrimalTangentPairs(np.ceil(self.primal), 0 * self.tangent)          # ... ceil, 0

    # NumPy logarithm and exponent operators

    def exp(self):
        # Exponentiate
        return PrimalTangentPairs(np.exp(self.primal),                    # Exponentiate primal
                                  np.exp(self.primal) * self.tangent)     # ... exponential rule

    def log(self):
        # Logarithm base-e
        return PrimalTangentPairs(np.log(self.primal),                    # Natural logarithm primal
                                  self.tangent / self.primal)             # ... logarithm rule

    def log2(self):
        # Logarithm base-2
        pair = self.log()                                                 # Compute natural log
        return PrimalTangentPairs(pair.primal / np.log(2),                # Apply logarithm rule for primal
                                  pair.tangent / np.log(2))               # ... and logarithm rule for tangent

    def log10(self):
        # Logarithm base-10
        pair = self.log()                                                 # Compute natural log
        return PrimalTangentPairs(pair.primal / np.log(10),               # Apply logarithm rule for primal
                                  pair.tangent / np.log(10))              # ... and logarithm rule for tangent

    def sqrt(self):
        # Square root
        return PrimalTangentPairs(np.sqrt(self.primal),                        # Compute square root of primal
                                  self.tangent / (2 * np.sqrt(self.primal)))   # ... power rule for tangent

    # NumPy trigonometry operators

    def sin(self):
        # Sine function
        return PrimalTangentPairs(np.sin(self.primal),
                                  np.cos(self.primal) * self.tangent)

    def cos(self):
        # Cosine function
        return PrimalTangentPairs(np.cos(self.primal),
                                  -np.sin(self.primal) * self.tangent)

    def tan(self):
        # Tangent function
        return PrimalTangentPairs(np.tan(self.primal),
                                  self.tangent / np.cos(self.primal)**2)

    def arcsin(self):
        # Inverse sine function
        return PrimalTangentPairs(np.arcsin(self.primal),
                                  self.tangent / np.sqrt(1 - self.primal**2))

    def arccos(self):
        # Inverse cosine function
        return PrimalTangentPairs(np.arccos(self.primal),
                                  -1 * self.tangent / np.sqrt(1 - self.primal**2))

    def arctan(self):
        # Inverse tangent function
        return PrimalTangentPairs(np.arctan(self.primal),
                                  self.tangent / (1 + self.primal**2))

    def sinh(self):
        # Hyperbolic sine function
        return PrimalTangentPairs(np.sinh(self.primal),
                                  np.cosh(self.primal) * self.tangent)

    def cosh(self):
        # Hyperbolic cosine function
        return PrimalTangentPairs(np.cosh(self.primal),
                                  np.sinh(self.primal) * self.tangent)

    def tanh(self):
        # Hyperbolic tangent function
        return PrimalTangentPairs(np.tanh(self.primal),
                                  self.tangent / (np.cosh(self.primal) ** 2))

    def arcsinh(self):
        # Inverse hyperbolic sine function
        return PrimalTangentPairs(np.arcsinh(self.primal),
                                  self.tangent / np.sqrt(self.primal**2 + 1))

    def arccosh(self):
        # Inverse hyperbolic cosine function
        return PrimalTangentPairs(np.arccosh(self.primal),
                                  self.tangent / np.sqrt(self.primal**2 - 1))

    def arctanh(self):
        # Inverse hyperbolic tangent function
        return PrimalTangentPairs(np.arctanh(self.primal),
                                  self.tangent / (1 - self.primal**2))

    # SciPy special operators

    def polygamma(self, n):
        # Polygamma function
        return PrimalTangentPairs(float(sp.special.polygamma(n, self.primal)),
                                  self.tangent * sp.special.polygamma(n+1, self.primal))

    def normal_cdf(self):
        # Cumulative Distribution Function (CDF) of the Normal Distribution
        return PrimalTangentPairs(float(norm.cdf(self.primal)),
                                  self.tangent * norm.pdf(self.primal))

    def normal_pdf(self):
        # Probability Density Function (PDF) of the Normal Distribution
        dx_pdf = -(np.exp(-self.primal**2 / 2) * self.primal) / np.sqrt(2 * np.pi)
        return PrimalTangentPairs(float(norm.pdf(self.primal)),
                                  self.tangent * dx_pdf)



#####################################################################################################################
# Functions to raise common errors across functionalities
#####################################################################################################################

import numpy as np


def check_alpha_level(alpha):
    """Error checking function to verify :math:`alpha` is between [0,1].

    Parameters
    ----------
    alpha : float
        User-provided alpha value

    Returns
    -------
    None
    """
    if not 0 < alpha < 1:
        raise ValueError("`alpha` must be 0 < a < 1")


def check_variance_is_not_none(variance):
    """Error checking function to verify variance was successfully estimated.

    Parameters
    ----------
    variance : float, ndarray, None
        Internally computed variance

    Returns
    -------
    None
    """
    if variance is None:
        raise ValueError("Either theta has not been estimated yet, or there is a np.nan in the bread matrix. "
                         "Therefore, confidence_intervals() cannot be called.")


def check_penalty_shape(theta, penalty, center):
    """Error checking function to verify provided penalty terms are valid.

    Parameters
    ----------
    theta : float, ndarray
        Parameters to be estimate for a regression model
    penalty : ndarray
        User-provided penalty terms
    center : ndarray
        User-provided place to penalize parameters towards

    Returns
    -------
    None
    """
    # Checking the penalty term is correct shape
    if penalty.size != 1:
        if penalty.shape[0] != len(theta):
            raise ValueError("The penalty term must be either a single number or the same length as theta.")

    # Checking the penalty center is correct shape
    if center.size != 1:
        if center.shape[0] != len(theta):
            raise ValueError("The center term must be either a single number or the same length as theta.")

    # Checking the penalty term is non-negative
    if np.any(penalty < 0):
        raise ValueError("All penalty terms must be non-negative")


def check_survival_data_valid(delta, time):
    """Error checking function to verify input survival data is valid.

    Parameters
    ----------
    delta : ndarray
        User-provided event indicator
    time : ndarray
        User-provided times

    Returns
    -------
    None
    """
    delta_no_nan = delta[~np.isnan(delta)]
    if not np.isin(delta_no_nan, [0, 1]).all():
        raise ValueError("All non-missing event indicator values must be either zero or one for survival models.")

    time_no_nan = time[~np.isnan(time)]
    if not np.where(time_no_nan > 0, 1, 0).all():
        raise ValueError("All non-missing observed times must be non-negative for survival models.")



#####################################################################################################################
# Functionality to compute the sandwich
#   This script allows for computation of the empirical sandwich variance estimator with just the
#   parameter values and estimating equations. This is to allow computing the sandwich quickly without
#   called the MEstimator procedure itself.
#####################################################################################################################

import warnings

import numpy as np
from scipy.optimize import approx_fprime
from scipy.stats import norm

from delicatessen.errors import check_alpha_level
from delicatessen.derivative import auto_differentiation, approx_differentiation


def compute_sandwich(stacked_equations, theta, deriv_method='approx', dx=1e-9, allow_pinv=True):
    r"""Compute the empirical sandwich variance estimator given a set of estimating equations and parameter estimates.
    Note that this functionality does not solve for the parameter estimates (unlike ``MEstimator``). Instead, it
    only computes the sandwich for the provided value.

    The empirical sandwich variance estimator is defined as

    .. math::

        V_n(O_i; \theta) = B_n(O_i; \theta)^{-1} F_n(O_i; \theta) \left[ B_n(O_i; \theta)^{-1} \right]^{T}

    where :math:`\psi(O_i; \theta)` is the estimating function,

    .. math::

        B_n(O_i; \theta) = \sum_{i=1}^n \frac{\partial}{\partial \theta} \psi(O_i; \theta),

    and

    .. math::

        F_n(O_i; \theta) = \sum_{i=1}^n \psi(O_i; \theta) \psi(O_i; \theta)^T .

    To compute the bread matrix, :math:`B_n`, the matrix of partial derivatives is computed by using either finite
    difference methods or automatic differentiation. For finite differences, the default is to use SciPy's
    ``approx_fprime`` functionality, which uses forward finite differences. However, you can also use the delicatessen
    homebrew version that allows for forward, backward, and center differences. Automatic differentiation is also
    supported by a homebrew version.

    To compute the meat matrix, :math:`F_n`, only linear algebra methods, implemented through NumPy, are necessary.
    The sandwich is then constructed from these pieces using linear algebra methods from NumPy.

    Parameters
    ----------
    stacked_equations : function, callable
        Function that returns a `v`-by-`n` NumPy array of the estimating equations. See provided examples in the
        documentation for how to construct a set of estimating equations.
    theta : list, set, array
        Parameter estimates to compute the empirical sandwich variance estimator at. Note that this function assumes
        that you have solved for the ``theta`` that correspond to the root of the input estimating equations.
    deriv_method : str, optional
        Method to compute the derivative of the estimating equations for the bread matrix. Options include numerical
        approximation via the forward difference method via SciPy (``'approx'``), forward difference implemented by-hand
        (`'fapprox'`), backward difference implemented by-hand (`'bapprox'`),  central difference implemented by-hand
        (`'capprox'`), or forward-mode automatic differentiation (``'exact'``). Default is ``'approx'``.
    dx : float, optional
        Spacing to use to numerically approximate the partial derivatives of the bread matrix. Here, a small value
        for ``dx`` should be used, since some large values can result in poor approximations. This argument is only
        used with numerical approximation methods. Default is ``1e-9``.
    allow_pinv : bool, optional
        Whether to allow for the pseudo-inverse (via ``numpy.linalg.pinv``) if the bread matrix is determined to be
        non-invertible. If you want to disallow the pseudo-inverse (i.e., use ``numpy.linalg.inv``), set this
        argument to ``False``. Default is ``True``, which  is more robust to the possible bread matrices.

    Returns
    -------
    array :
        Returns a `p`-by-`p` NumPy array for the input ``theta``, where ``p = len(theta)``

    Examples
    --------
    Loading necessary functions and building a generic data set for estimation of the mean

    >>> import numpy as np
    >>> from delicatessen import MEstimator
    >>> from delicatessen import compute_sandwich
    >>> from delicatessen.estimating_equations import ee_mean_variance

    >>> y_dat = [1, 2, 4, 1, 2, 3, 1, 5, 2]

    The following is an illustration of how to compute sandwich covariance using only an estimating equation and the
    parameter values. The mean and variance (that correspond to ``ee_mean_variance``) can be computed using NumPy by

    >>> mean = np.mean(y_dat)
    >>> var = np.var(y_dat, ddof=0)

    For the corresponding estimating equation, we can use the built-in functionality as done below

    >>> def psi(theta):
    >>>     return ee_mean_variance(theta=theta, y=y_dat)

    Calling the sandwich computation procedure

    >>> sandwich_asymp = compute_sandwich(stacked_equations=psi, theta=[mean, var])

    The output sandwich is the *asymptotic* variance (or the variance that corresponds to the standard deviation). To
    get the variance (or the variance that corresponds to the standard error), we rescale ``sandwich`` by the number of
    observations

    >>> sandwich = sandwich_asymp / len(y_dat)

    The standard errors are then

    >>> se = np.sqrt(np.diag(sandwich))

    References
    ----------
    Boos DD, & Stefanski LA. (2013). M-estimation (estimating equations). In Essential Statistical Inference
    (pp. 297-337). Springer, New York, NY.

    Ross RK, Zivich PN, Stringer JSA, & Cole SR. (2024). M-estimation for common epidemiological measures: introduction
    and applied examples. *International Journal of Epidemiology*, 53(2), dyae030.

    Stefanski LA, & Boos DD. (2002). The calculus of M-estimation. The American Statistician, 56(1), 29-38.
    """
    # Evaluating at provided theta values
    evald_theta = np.asarray(stacked_equations(theta=theta))        # Evaluating EE at theta-hat
    if len(theta) == 1:                                             # Number of parameters
        n_obs = evald_theta.shape[0]                                # ... to get number of obs
    else:                                                           # Number of parameters
        n_obs = evald_theta.shape[1]                                # ... to get number of obs

    # Step 1: Compute the bread matrix
    bread = compute_bread(stacked_equations=stacked_equations,      # Call the bread matrix function
                          theta=theta,                              # ... at given theta-hat
                          deriv_method=deriv_method,                # ... with derivative method
                          dx=dx)                                    # ... and approximation
    bread = bread / n_obs                                           # Scale bread by number of obs

    # Step 2: Compute the meat matrix
    meat = compute_meat(stacked_equations=stacked_equations,        # Call the meat matrix function
                        theta=theta)                                # ... at given theta-hat
    meat = meat / n_obs                                             # Scale meat by number of obs

    # Step 3: Construct sandwich from the bread and meat matrices
    sandwich = build_sandwich(bread=bread,                          # Call the sandwich constructor
                              meat=meat,                            # ... with bread and meat matrices above
                              allow_pinv=allow_pinv)                # ... and whether to allow pinv

    # Return the constructed empirical sandwich variance estimator
    return sandwich


def compute_bread(stacked_equations, theta, deriv_method, dx=1e-9):
    r"""Function to compute the bread matrix. The bread matrix is defined as

    .. math::

        B_n(O_i; \theta) = \sum_{i=1}^n \frac{\partial }{\partial \theta} \psi(O_i; \theta)

    where :math:`\psi(O_i; \theta)` is the estimating function.
    To compute the bread matrix, :math:`B_n`, the matrix of partial derivatives is computed by using either finite
    difference methods or automatic differentiation. For finite differences, the default is to use SciPy's
    ``approx_fprime`` functionality, which uses forward finite differences. However, you can also use the delicatessen
    homebrew version that allows for forward, backward, and center differences. Automatic differentiation is also
    supported by a homebrew version.

    Parameters
    ----------
    stacked_equations : function, callable
        Function that returns a `v`-by-`n` NumPy array of the estimating equations. See provided examples in the
        documentation for how to construct a set of estimating equations.
    theta : list, set, array
        Parameter estimates to compute the empirical sandwich variance estimator at. Note that this function assumes
        that you have solved for the ``theta`` that correspond to the root of the input estimating equations.
    deriv_method : str, optional
        Method to compute the derivative of the estimating equations for the bread matrix. Options include numerical
        approximation via the forward difference method via SciPy (``'approx'``), forward difference implemented by-hand
        (`'fapprox'`), backward difference implemented by-hand (`'bapprox'`),  central difference implemented by-hand
        (`'capprox'`), or forward-mode automatic differentiation (``'exact'``). Default is ``'approx'``.
    dx : float, optional
        Spacing to use to numerically approximate the partial derivatives of the bread matrix. Here, a small value
        for ``dx`` should be used, since some large values can result in poor approximations. This argument is only
        used when numerical approximation methods. Default is ``1e-9``.

    Returns
    -------
    array :
        Returns a `p`-by-`p` NumPy array for the input ``theta``, where ``p = len(theta)``
    """
    def estimating_equation(input_theta):
        ef = np.asarray(stacked_equations(theta=input_theta))
        if ef.ndim == 1:
            return np.sum(ef)
        else:
            return np.sum(ef, axis=1)

    # Compute the derivative
    if deriv_method.lower() == 'approx':
        bread_matrix = approx_fprime(xk=theta,
                                     f=estimating_equation,
                                     epsilon=dx)
        if bread_matrix.ndim == 1:
            bread_matrix = np.asarray([bread_matrix, ])
    elif deriv_method.lower() == 'capprox':
        bread_matrix = approx_differentiation(xk=theta,
                                              f=estimating_equation,
                                              method='capprox',
                                              epsilon=dx)
    elif deriv_method.lower() == 'fapprox':
        bread_matrix = approx_differentiation(xk=theta,
                                              f=estimating_equation,
                                              method='fapprox',
                                              epsilon=dx)
    elif deriv_method.lower() == 'bapprox':
        bread_matrix = approx_differentiation(xk=theta,
                                              f=estimating_equation,
                                              method='bapprox',
                                              epsilon=dx)
    elif deriv_method.lower() == "exact":  # Automatic Differentiation
        bread_matrix = auto_differentiation(xk=theta,  # Compute the exact derivative at theta
                                            f=estimating_equation)  # ... for the given estimating equations
    else:
        raise ValueError("The input for deriv_method was "
                         + str(deriv_method)
                         + ", but only 'approx', 'fapprox', 'capprox', 'bapprox' "
                           "and 'exact' are available.")

    # Checking for an issue when trying to invert the bread matrix
    if np.isnan(bread_matrix).any():
        warnings.warn("The bread matrix contains at least one np.nan, so it cannot be inverted. The variance will "
                      "not be calculated. This may be an issue with the provided estimating equations or the "
                      "evaluated theta.",
                      UserWarning)

    # Returning the constructed bread matrix according to SB 2002
    return -1 * bread_matrix


def compute_meat(stacked_equations, theta):
    r"""Function to compute the meat matrix. The meat matrix is defined as

    .. math::

        F_n(O_i; \theta) = \sum_{i=1}^n \psi(O_i; \theta) \psi(O_i; \theta)^T

    where :math:`\psi(O_i; \theta)` is the estimating function.
    Rather than summing over all the individual contributions, this implementation takes a single dot product of the
    stacked estimating functions. This implementation is much faster than summing over :math:`n` matrices.

    Parameters
    ----------
    stacked_equations : function, callable
        Function that returns a `v`-by-`n` NumPy array of the estimating equations. See provided examples in the
        documentation for how to construct a set of estimating equations.
    theta : list, set, array
        Parameter estimates to compute the empirical sandwich variance estimator at. Note that this function assumes
        that you have solved for the ``theta`` that correspond to the root of the input estimating equations.

    Returns
    -------
    array :
        Returns a `p`-by-`p` NumPy array for the input ``theta``, where ``p = len(theta)``
    """
    evald_theta = np.asarray(stacked_equations(theta=theta))  # Evaluating EE at theta-hat
    return np.dot(evald_theta, evald_theta.T)                 # Return the fast dot product calculation


def build_sandwich(bread, meat, allow_pinv=True):
    r"""Function to combine the sandwich elements together. This function takes the bread and meat matrices, does the
    inversions, and then combines them together. This function is separate from ``compute_sandwich`` as it is called
    by both ``compute_sandwich`` and ``MEstimator``.

    Parameters
    ----------
    bread : ndarray
        The bread matrix. The expected input is the output from the ``compute_bread`` function
    meat : ndarray
        The meat matrix. The expected input is the output from the ``compute_meat`` function
    allow_pinv : bool, optional
        Whether to allow for the pseudo-inverse (via ``numpy.linalg.pinv``) if the bread matrix is determined to be
        non-invertible. If you want to disallow the pseudo-inverse (i.e., use ``numpy.linalg.inv``), set this
        argument to ``False``. Default is ``True``, which  is more robust to the possible bread matrices.

    Returns
    -------
    array :
        Returns a `p`-by-`p` NumPy array for the input ``theta``, where ``p = len(theta)``
    """
    # Check if there is an issue with the bread matrix
    if np.any(np.isnan(bread)):                                   # If bread contains NaN, breaks
        return None                                               # ... so give back a NaN

    # Compute the bread inversion
    if allow_pinv:                                                 # Allowing the pseudo-inverse
        bread_invert = np.linalg.pinv(bread)                       # ... then call pinv
    else:                                                          # Only allowing the actual inverse
        bread_invert = np.linalg.inv(bread)                        # ... then call inv

    # Compute the sandwich variance
    sandwich = np.dot(np.dot(bread_invert, meat), bread_invert.T)

    # Return the sandwich covariance matrix
    return sandwich


def delta_method(theta, g, covariance, deriv_method='exact', dx=1e-9):
    r"""Function to apply the Delta Method for a given parameter vector and transformation. The Delta Method is defined
    as

    .. math::

        Var \left[ g(\theta) \right] \approx g'(\theta) \; \Sigma_{\theta} \; g'(\theta)^T

    where :math:`\theta` is the parameter vector, :math:`g` is a vector-valued function that returns a 1 dimensional
    vector, :math:`g'` is the gradient (or partial derivatives), and :math:`\Sigma_{\theta}` is the covariance matrix
    for :math:`\theta`. In words, the variance of the transformation of the parameters is equal to covariance of the
    parameters sandwiched between the derivatives of the transformation functions. This expression then provides a
    variance estimator when replacing :math:`\theta` and :math:`\Sigma` with :math:`\hat{\theta}` and
    :math:`\hat{\Sigma}`, respectively.

    As described elsewhere, the sandwich variance estimator automates the Delta Method. Therefore, one can simply
    program the corresponding estimating equation to estimate the variance for that transformation. However, this can be
    computationally inefficient when :math:`g` outputs a large vector. This functionality offers a way to apply the
    Delta Method outside of ``MEstimator`` and ``GMMEstimator``. Internally, this function is used for some prediction
    functionalities.

    Parameters
    ----------
    theta : ndarray, list, set
        Parameter vector of dimension `v` to apply the transformation function ``g`` with.
    g : function, callable
        Vector function that transforms the `v` dimension parameter vector ``theta`` into a `w` dimensional vector.
    covariance : ndarray, list, set
        Covariance matrix for the parameter vector ``x``.
    deriv_method : str, optional
        Method to compute the derivative of the function ``g``. Default is ``'exact'``. Options include numerical
        approximation via the forward difference method via SciPy (``'approx'``), forward difference implemented by-hand
        (`'fapprox'`), backward difference implemented by-hand (`'bapprox'`),  central difference implemented by-hand
        (`'capprox'`), or forward-mode automatic differentiation (``'exact'``).
    dx : float, optional
        Spacing to use to numerically approximate the partial derivatives of the bread matrix. Here, a small value
        for ``dx`` should be used, since some large values can result in poor approximations. This argument is only
        used when numerical approximation methods. Default is ``1e-9``.

    Returns
    -------
    array :
        Returns a `p`-by-`p` NumPy array for the input ``theta``, where `p` is the length of the output vector from the
        function :math:`g`

    Examples
    --------
    Using ``delta_method`` to compute the variance should be done similar to the following

    >>> import numpy as np
    >>> import pandas as pd
    >>> from delicatessen import MEstimator, delta_method
    >>> from delicatessen.estimating_equations import ee_ipw

    To illustrate how ``delta_method`` is intended to be used, we will first use an M-estimator to compute the point
    and variance estimates for the parameter vector :math:`\theta`. Here, we will replicate the example from the
    documentation for ``ee_ipw``.

    >>> n = 200
    >>> d = pd.DataFrame()
    >>> d['W'] = np.random.binomial(1, p=0.5, size=n)
    >>> d['A'] = np.random.binomial(1, p=(0.25 + 0.5*d['W']), size=n)
    >>> d['Ya0'] = np.random.binomial(1, p=(0.75 - 0.5*d['W']), size=n)
    >>> d['Ya1'] = np.random.binomial(1, p=(0.75 - 0.5*d['W'] - 0.1*1), size=n)
    >>> d['Y'] = (1-d['A'])*d['Ya0'] + d['A']*d['Ya1']
    >>> d['C'] = 1

    Defining psi, or the stacked estimating equations

    >>> def psi(theta):
    >>>     return ee_ipw(theta, y=d['Y'], A=d['A'],
    >>>                   W=d[['C', 'W']])

    Calling the M-estimation procedure

    >>> estr = MEstimator(stacked_equations=psi, init=[0., 0.5, 0.5, 0., 0.])
    >>> estr.estimate(solver='lm')

    Per the documentation for ``ee_ipw``, the average causal effect (or risk difference) is given by the following

    >>> estr.theta[0]        # causal mean difference of 1 versus 0
    >>> estr.variance[0, 0]  # corresponding variance estimate

    Suppose that ``ee_ipw`` did not directly provide the risk difference. Further, imagine we were interested in the
    risk ratio (log-transformed). Both of these quantities are transformations of the risk under action 1
    (i.e., ``theta[1]``) and the risk under action 0 (i.e., ``theta[2]``). The following is a function that applies and
    returns those transformations

    >>> def causal_contrasts(theta):
    >>>     risk1, risk0 = theta
    >>>     risk_diff = risk1 - risk0
    >>>     log_risk_ratio = np.log(risk1) - np.log(risk0)
    >>>     return risk_diff, log_risk_ratio

    To estimate the variance for this transformation, one can now use the *delta method*. While one could manually
    compute the derivatives for this function, ``delta_method`` automates this procedure for you (using either automatic
    or numerical approximation methods). Below is how ``delta_method`` can be applied to compute the variance for the
    risk difference and risk ratio

    >>> risks = estr.theta[1:3]                     # Risks
    >>> risks_c0var = estr.variance[1:3, 1:3]       # Variance for risks
    >>> rd, log_rr = causal_contrasts(theta=risks)  # RD, log(RR)

    >>> covar = delta_method(theta=risks,             # Delta method
    >>>                      g=causal_contrasts,      # ... function g
    >>>                      covariance=risks_c0var)  # ... covariance

    The output from ``delta_method`` is the corresponding covariance matrix for risk difference and risk ratio. We can
    get the corresponding 95% confidence intervals via

    >>> rd_stderr = np.sqrt(covar[0, 0])
    >>> rd_lcl, rd_ucl = rd - 1.96*rd_stderr, rd + 1.96*rd_stderr
    >>> rr_stderr = np.sqrt(covar[1, 1])
    >>> rr_lcl, rr_ucl = np.exp(log_rr - 1.96*rr_stderr), np.exp(log_rr + 1.96*rr_stderr)

    While these transformations are straightforward to stack as estimating functions, ``delta_method`` offers another
    option for estimating the variance of transformations that has computational benefits in some settings.

    References
    ----------
    Boos DD, & Stefanski LA. (2013). Large Sample Theory: The Basics, In
    *Essential Statistical Inference: Theory and Methods*, 237-240.

    Cox C. (2005). Delta method. *Encyclopedia of Biostatistics*.

    Oehlert GW. (1992). A Note on the Delta Method. *The American Statistician*, 46(1), 27-29.
    """
    # Setup
    covariance = np.asarray(covariance)
    j = covariance.shape[0]
    k = covariance.shape[1]

    def g_array(theta):
        # Function to convert user function g to always return a NumPy array
        return np.asarray(g(theta))

    # Computing g(\theta)
    theta_star = g_array(theta)
    theta_dim = theta_star.shape
    v = len(theta)

    # Checking dimensions of parameter vector
    if len(theta_dim) > 1:
        raise ValueError("Output from function `g` must be a one-dimensional array")
    if j != k:
        raise ValueError("Input covariance matrix must be symmetric. Input matrix had dimensions " + str(j)
                         + " rows and " + str(k) + " columns")
    if j != v:
        raise ValueError("Input parameter vector and covariance matrix must share a dimension, but theta has " + str(v)
                         + " as its dimension and the matrix had dimension " + str(j))

    # Computing derivative for g(x)
    if deriv_method.lower() == "exact":
        g_prime = auto_differentiation(xk=theta, f=g_array)
    elif deriv_method.lower() == 'approx':
        g_prime = approx_fprime(xk=theta, f=g_array, epsilon=dx)
        if g_prime.ndim == 1:
            g_prime = np.asarray([g_prime, ])
    elif deriv_method.lower() in ['capprox', 'fapprox', 'bapprox']:
        g_prime = approx_differentiation(xk=theta, f=g_array, method=deriv_method.lower(), epsilon=dx)
    else:
        raise ValueError("The input for deriv_method was " + str(deriv_method)
                         + ", but only 'approx', 'fapprox', 'capprox', 'bapprox' and 'exact' are available.")

    # Applying delta method to get the variance
    covariance_g = np.dot(np.dot(g_prime, covariance), g_prime.T)

    # Returning the covariance matrix for g(x)
    return covariance_g


def compute_critical_value_bands(theta, covariance, alpha=0.05, method='supt', n_draws=100000, seed=None):
    """Function to compute the critical value for parameter vectors.

    Confidence bands are an extension of confidence intervals. Confidence intervals claim to cover the true parameter
    under infinite repetitions at the :math:`1-\alpha` proportion. However, when considering vectors of parameters,
    confidence interval coverage is below this proportion (can be thought of as a multiple-testing problem). Confidence
    bands instead work to ensure :math:`1-\alpha` coverage of *the parameter vector*. Many of the algorithms for
    confidence bands operate by adjusting the critical value. This function offers several methods to compute the
    critical value for confidence bands

    Parameters
    ----------
    theta : ndarray, list, set
        Parameter vector of dimension `v` to compute confidence bands for. This vector should consist of all parameters
        that inference is being jointly drawn for.
    covariance : ndarray, list, set
        Covariance matrix of dimension `v`. This should be the empirical sandwich covariance matrix for the parameter.
    alpha : float, optional
        The :math:`0 < \alpha < 1` level for the corresponding confidence bands. Default is 0.05, which
        corresponds to 95% confidence bands.
    method : str, optional
        Method to compute the confidence bands. Currently, only the sup-t and Bonferroni method are supported.
        Default is ``'supt'``
    n_draws : int, optional
        Number of random draws to use for any methods based on simulated approximation. Default is ``100000``.
    seed : int, optional
        Seed to intialize a pseudo RNG for methods based on simulated approximations. Default is ``None``
        which does not use a reproducible seed. To consistently obtain the exact same confidence bands, please use
        a seed.

    Returns
    -------
    float :
        Corresponding critical value for given method and inputs

    References
    ----------
    Montiel Olea JL & Plagborg‐Møller M. (2019). Simultaneous confidence bands: Theory, implementation, and an
    application to SVARs. *Journal of Applied Econometrics*, 34(1), 1-17.

    Zivich PN, Cole SR, Greifer N, Montoya LM, Kosorok MR, Edwards JK. (2025). Confidence Regions for Multiple
    Outcomes, Effect Modifiers, and Other Multiple Comparisons, *arXiv:2510.07076*
    """
    # Pre-processing inputs
    theta = np.asarray(theta)
    covariance = np.asarray(covariance)

    # Error checking for inputs
    if len(theta.shape) != 1:
        raise ValueError("The input theta vector must be a 1-dimensional vector. Instead, the input theta vector "
                         "has a dimension of " + str(len(theta.shape)))
    if len(covariance.shape) != 2:
        raise ValueError("The input covariance matrix must be a 2-dimensional matrix. Instead, the input covariance "
                         "matrix has a dimension of " + str(len(covariance.shape)))
    if covariance.shape[0] != covariance.shape[1]:
        raise ValueError("The input covariance matrix must be square (i.e., have the same number of rows and columns). "
                         "Instead, the input covariance matrix has " + str(covariance.shape[0]) + " rows and "
                         + str(covariance.shape[1]) + " columns.")
    if theta.shape[0] != covariance.shape[0]:
        raise ValueError("The dimension of the theta vector and covariance matrix must be equal. Instead, the theta "
                         "vector has " + str(theta.shape[0]) + " elements and the covariance matrix is dimension "
                         + str(covariance.shape[0]))
    check_alpha_level(alpha=alpha)

    # Processing
    stderr = np.diag(covariance) ** 0.5

    # Approximate c
    rng = np.random.default_rng(seed=seed)
    k = len(theta)
    if method.lower() in ['supt', 'sup-t']:
        mvn = rng.multivariate_normal([0., ] * k, cov=covariance, size=n_draws)
        if (stderr <= 0).any():
            raise ValueError("There is at least one parameter with a standard error of zero or less. The sup-t method "
                             "cannot be applied as it would require division by zero for the rescaling process.")
        scaled_mvn = np.abs(mvn / stderr)
        ts = np.max(scaled_mvn, axis=1)
        critical_value = np.percentile(ts, q=(1 - alpha) * 100)
    elif method.lower() == 'bonferroni':
        critical_value = norm.ppf(1 - alpha / (2 * k), loc=0, scale=1)
    else:
        raise ValueError("The method '" + str(method) + "' was specified, but only the following "
                         "methods are supported: 'supt', 'bonferroni'.")

    # Returning the corresponding critical value
    return critical_value


def compute_confidence_bands(theta, covariance, alpha=0.05, method='supt', n_draws=100000, seed=None):
    r"""Function to compute the confidence bands for a given parameter vector and covariance matrix.

    Confidence bands are an extension of confidence intervals. Confidence intervals claim to cover the true parameter
    under infinite repetitions at the :math:`1-\alpha` proportion. However, when considering vectors of parameters,
    confidence interval coverage is below this proportion (can be thought of as a multiple-testing problem). Confidence
    bands instead work to ensure :math:`1-\alpha` coverage of *the parameter vector*. Many of the algorithms for
    confidence bands operate by adjusting the critical value.

    Parameters
    ----------
    theta : ndarray, list, set
        Parameter vector of dimension `v` to compute confidence bands for. This vector should consist of all parameters
        that inference is being jointly drawn for.
    covariance : ndarray, list, set
        Covariance matrix of dimension `v`. This should be the empirical sandwich covariance matrix for the parameter.
    alpha : float, optional
        The :math:`0 < \alpha < 1` level for the corresponding confidence bands. Default is 0.05, which
        corresponds to 95% confidence bands.
    method : str, optional
        Method to compute the confidence bands. Currently, only the sup-t and Bonferroni method are supported.
        Default is ``'supt'``
    n_draws : int, optional
        Number of random draws to use for any methods based on simulated approximation. Default is ``100000``.
    seed : int, optional
        Seed to intialize a pseudo RNG for methods based on simulated approximations. Default is ``None``
        which does not use a reproducible seed. To consistently obtain the exact same confidence bands, please use
        a seed.

    Returns
    -------
    array :
        b-by-2 array, where row 1 is the confidence intervals for :math:`\theta_1`, ..., and row b is the confidence
        intervals for :math:`\theta_b`

    Examples
    --------

    References
    ----------
    Montiel Olea JL & Plagborg‐Møller M. (2019). Simultaneous confidence bands: Theory, implementation, and an
    application to SVARs. *Journal of Applied Econometrics*, 34(1), 1-17.

    Zivich PN, Cole SR, Greifer N, Montoya LM, Kosorok MR, Edwards JK. (2025). Confidence Regions for Multiple
    Outcomes, Effect Modifiers, and Other Multiple Comparisons, *arXiv:2510.07076*
    """
    # Computing the critical value
    critical_value = compute_critical_value_bands(theta=theta, covariance=covariance,
                                                  alpha=alpha, method=method,
                                                  n_draws=n_draws, seed=seed)

    # Processing inputs
    theta = np.asarray(theta)
    covariance = np.asarray(covariance)
    stderr = np.diag(covariance) ** 0.5

    # Computing the corresponding confidence bands
    conf_bands = np.asarray([theta - critical_value * stderr,
                             theta + critical_value * stderr])

    # Returning confidence bands
    return conf_bands.T



#####################################################################################################################
# Implementation of the M-estimator
#####################################################################################################################

import warnings
import numpy as np
from scipy.optimize import newton, root, minimize
from scipy.stats import norm
from scipy.optimize import OptimizeWarning

from delicatessen.errors import check_alpha_level, check_variance_is_not_none
from delicatessen.sandwich import compute_bread, compute_meat, build_sandwich, compute_confidence_bands


class _GeneralEstimator:
    """Internal base class for M-estimator and GMM-estimator to be built from. Not intended for users.
    """
    def __init__(self, stacked_equations, init, subset=None):
        self.stacked_equations = stacked_equations     # User-input stacked estimating equations
        self.init = init                               # User-input initial starting values for solving estimating eqs
        if subset is None:                             # Handling subset of parameters
            self._subset_ = subset                     # ... when None, set as None
        else:                                          # Otherwise
            self._subset_ = sorted(subset)             # ... ensure it is sorted

        # Storage for results from the Estimation procedure
        self.n_obs = None                 # Number of unique observations (calculated later)
        self.theta = None                 # Optimized theta values (calculated later)
        self.bread = None                 # Bread from theta values (calculated later)
        self.meat = None                  # Meat from theta values (calculated later)
        self.variance = None              # Covariance matrix for theta values (calculated later)
        self.asymptotic_variance = None   # Asymptotic covariance matrix for theta values (calculated later)
        self.weight_matrix = None         # Weight matrix for the GMM estimator

    def confidence_intervals(self, alpha=0.05):
        r"""Calculate two-sided Wald-type :math:`(1 - \alpha) \times` 100% confidence intervals using the point
        and sandwich variance estimates. The formula for the confidence intervals is

        .. math::

            \hat{\theta} \pm z_{\alpha / 2} \times \widehat{SE}(\hat{\theta})

        Note
        ----
        The ``.estimate()`` function must be called before the confidence intervals can be calculated.

        Parameters
        ----------
        alpha : float, optional
            The :math:`0 < \alpha < 1` level for the corresponding confidence intervals. Default is 0.05, which
            corresponds to 95% confidence intervals.

        Returns
        -------
        array :
            b-by-2 array, where row 1 is the confidence intervals for :math:`\theta_1`, ..., and row b is the confidence
            intervals for :math:`\theta_b`
        """
        # Check that variance has been estimated
        check_variance_is_not_none(variance=self.variance)

        # Check valid alpha value is being provided
        check_alpha_level(alpha=alpha)

        # 'Looking up' via Z table
        z_alpha = norm.ppf(1 - alpha / 2, loc=0, scale=1)   # Z_alpha value for CI

        # Calculating confidence intervals
        param_se = np.sqrt(np.diag(self.variance))          # Take the diagonal of the sandwich and then SQRT
        lower_ci = self.theta - z_alpha * param_se          # Calculate lower CI
        upper_ci = self.theta + z_alpha * param_se          # Calculate upper CI

        # Return 2D array of lower and upper confidence intervals
        return np.asarray([lower_ci, upper_ci]).T

    def z_scores(self, null=0):
        r"""Calculate the Z-score using the point estimates and the sandwich variance. The formula for the Z score is

        .. math::

            \frac{\hat{\theta} - \theta}{\widehat{SE}(\hat{\theta})}

        where :math:`\theta` is the null. The ``.estimate()`` function must be called before the Z-scores can be
        calculated. Note that the default value for the null is zero.

        Parameters
        ----------
        null: int, float, ndarray, optional
            Null or reference for the the corresponding P-values. Default is 0.

        Returns
        -------
        array :
            Array of Z-scores for :math:`\theta_1, ..., \theta_b`, respectively
        """
        # Check that variance has been estimated
        check_variance_is_not_none(variance=self.variance)

        # Calculating Z-scores
        se = np.sqrt(np.diag(self.variance))       # Extract the standard error estimates from the sandwich
        z_score = (self.theta - null) / se         # Compute the Z-score
        return z_score                             # Return the Z-score to the user

    def p_values(self, null=0):
        r"""Calculate two-sided Wald-type P-values using the Z-scores compute using the point and the sandwich variance
        estimates. Once the Z-scores are computed, the corresponding P-values are obtained by comparing to the standard
        normal distribution.

        The ``.estimate()`` function must be called before the P-values can be calculated.

        Parameters
        ----------
        null: int, float, ndarray, optional
            Null or reference for the the corresponding P-values. Default is 0.

        Returns
        -------
        array :
            Array of P-values for :math:`\theta_1, ..., \theta_b`, respectively
        """
        z_score = self.z_scores(null=null)         # Calculating the Z-scores
        p_value = norm.sf(np.abs(z_score)) * 2     # Compute the corresponding P-values
        return p_value                             # Return P-values to the user

    def s_values(self, null=0):
        r"""Calculate two-sided Wald-type S-values using the point estimates and the sandwich variance. The S-value,
        or Shannon Information value, is a transformation of the P-values that has been argued to be more easily
        interpretable as it can be related back to simple coin-flipping scenarios. The transformation from a P-value
        into a S-value is.

        .. math::

            S = - \log_2(P)

        where :math:`P` is the corresponding P-value. The ``.estimate()`` function must be called before the S-values
        can be calculated.

        The S-value can be contextualized in terms of coin-flips. Suppose the S-value is :math:`s`. Then :math:`s`
        corresponds to the number of heads in a row with a fair coin (equal chances heads or tails). As :math:`s`
        increases, one would be more 'surprised' by the result (e.g., it might not be surprising to have 2 heads in a
        row, but it would be surprising for 20 in a row).

        Parameters
        ----------
        null: int, float, ndarray, optional
            Null or reference for the the corresponding S-values. Default is 0.

        Returns
        -------
        array :
            Array of S-values for :math:`\theta_1, ..., \theta_b`, respectively

        References
        ----------
        Cole SR, Edwards JK, & Greenland S. (2021). Surprise! *American Journal of Epidemiology*, 190(2), 191-193.

        Greenland S. (2019). Valid P-values behave exactly as they should: Some misleading criticisms of P-values and
        their resolution with S-values. *The American Statistician*, 73(sup1), 106-114.
        """
        p_values = self.p_values(null=null)          # Calculate P-values
        s_values = -1 * np.log2(p_values)            # Transform into S-values
        return s_values                              # Return S-values to the user

    def influence_functions(self, allow_pinv=True):
        r"""Calculate the influence function values for individual observations. Note that the influence function for
        :math:`i` is computed using the following equality

        .. math::

            IF(O_i; \theta) = \left\{ B_n(theta) \right\}^{-1} \psi(O_i; \theta)

        or that the influence functions is equal to the estimating function scaled by the inverse of the bread matrix.
        While this approach was historically used to derive the influence function, there are more general alternatives.
        `delicatessen` offers this implementation since

        Returns
        -------
        array :
            Returns a `n`-by-`b` NumPy array evaluated for the input ``theta``.

        References
        ----------
        Fisher A, & Kennedy EH. (2021). Visually communicating and teaching intuition for influence functions.
        *The American Statistician*, 75(2), 162-172.

        Hines O, Dukes O, Diaz-Ordaz K, & Vansteelandt S. (2022). Demystifying statistical learning based on efficient
        influence functions. *The American Statistician*, 76(3), 292-304.

        Stefanski LA, & Boos DD. (2002). The calculus of M-estimation. *The American Statistician*, 56(1), 29-38.
        """
        # Check that variance has been estimated
        check_variance_is_not_none(variance=self.variance)

        # Calculating estimating function
        efunc_i = self.stacked_equations(theta=self.theta)   # Computing estimating function at theta-hat
        if efunc_i.ndim == 1:                                # Checking if one dimension
            efunc_i = np.asarray([efunc_i, ])                # ... so can make 2D for dot product

        # Calculating inverse of bread matrix
        if allow_pinv:                                       # Allowing the pseudo-inverse
            bread_invert = np.linalg.pinv(self.bread)        # ... then call pinv
        else:                                                # Only allowing the actual inverse
            bread_invert = np.linalg.inv(self.bread)         # ... then call inv

        # Calculating influence function
        ifunc_i = np.dot(bread_invert, efunc_i)
        return ifunc_i.T

    def confidence_bands(self, subset=None, alpha=0.05, method='supt', n_draws=1000000, seed=None):
        r"""Calculate two-sided :math:`(1 - \alpha) \times` 100% confidence bands from the point and sandwich variance
        estimates. Rather than cover a single parameter, the confidence bands provide coverage for parameter *vectors*.
        The formula for the confidence bands is

        .. math::

            \hat{\theta} \pm \hat{c}_{\alpha / 2} \times \widehat{SE}(\hat{\theta})

        Note
        ----
        The ``.estimate()`` function must be called before the confidence bands can be calculated.


        This formula looks very similar to the confidence interval formula, but there is a different critical value.
        Note that the formula above has a hat on the critical value, :math:`c`, to denote it is being estimated.
        Currently, only the sup-t method for estimating the critical value is supported.

        Parameters
        ----------
        subset : list, set, array, None, optional
            Optional argument to compute the confidence bands for a subset of parameters. Note that this argument is
            distinct from the overall ``subset`` argument and is only used for the confidence bands. Default is
            ``None``, which computes the confidence bands for all parameters (as if they are are all the parameters of
            interest).
        alpha : float, optional
            The :math:`0 < \alpha < 1` level for the corresponding confidence bands. Default is 0.05, which
            corresponds to 95% confidence bands.
        method : str, optional
            Method to compute the confidence bands. Currently, only the sup-t and Bonferroni method are supported.
            Default is ``'supt'``
        n_draws : int, optional
            Number of random draws to use for any methods based on simulated approximation. Default is one million,
            ``1000000``.
        seed : int, optional
            Seed to initialize a pseudo RNG for methods based on simulated approximations. Default is ``None``
            which does not use a reproducible seed. To consistently obtain the same confidence bands with approximation
            methods, please use a seed.

        Returns
        -------
        array :
            b-by-2 array, where row 1 is the confidence intervals for :math:`\theta_1`, ..., and row b is the confidence
            intervals for :math:`\theta_b`

        References
        ----------
        Montiel Olea JL & Plagborg‐Møller M. (2019). Simultaneous confidence bands: Theory, implementation, and an
        application to SVARs. *Journal of Applied Econometrics*, 34(1), 1-17.
        """
        # Check that variance has been estimated
        check_variance_is_not_none(variance=self.variance)

        # Check valid alpha value is being provided
        check_alpha_level(alpha=alpha)

        # Processing parameter vector
        if subset is None:
            theta = self.theta
            covar = self.variance
        else:
            theta = self.theta[subset]
            covar = self.variance[np.ix_(subset, subset)]

        # Calculating confidence bands
        cb = compute_confidence_bands(theta=theta, covariance=covar,
                                      alpha=alpha,
                                      method=method,
                                      n_draws=n_draws, seed=seed)

        # Return 2D array of lower and upper confidence bands
        return cb

    @staticmethod
    def _eval_ee_(stacked_equations, subset):
        """Function to evaluate the estimating equations over the :math:`n` units for a given set of parameters

        Note
        ----
        Added in v1.0 to replace the inner functionality of ``_mestimation_answer_`` for the new ``approx_fprime`` but
        still support ``subset`` (without having to flip the subset flag).

        Parameters
        ----------
        stacked_equations :
            Estimating equations to evaluate
        subset :
            Whether to consider a subset of parameters

        Returns
        -------
        numpy.array
        """
        # IF stacked_equation returns 1 value, only return that 1 value
        if len(stacked_equations.shape) == 1:          # Checking length
            vals = np.sum(stacked_equations)           # ... avoid SciPy error by returning value rather than tuple
        # ELSE need to return a tuple for the root-finding procedure
        else:                                          # ... also considering how subset argument is handled
            # NOTE: switching to np.sum(..., axis=1) didn't speed things up versus a for-loop
            vals = ()                                  # ... create empty tuple
            rows = stacked_equations.shape[0]          # ... determine how many rows / parameters are present
            if subset is None:                         # ... if no subset, then simple loop where
                for i in range(rows):                  # ... go through each individual theta in the stack
                    row = stacked_equations[i, :]      # ... extract corresponding row
                    vals += (np.sum(row), )            # ... then add the theta sum to the tuple of thetas
            else:                                      # ... if subset, then conditional loop (to speed up)
                for i in range(rows):                  # ... go through each individual theta in the stack
                    if i in subset:                    # ... if parameter is in subset then
                        row = stacked_equations[i, :]  # ... extract corresponding row
                        vals += (np.sum(row), )        # ... then add the theta sum to the tuple of thetas
            vals = np.asarray(vals)                    # ... converting to a NumPy array for ease

        # Return the calculated values of theta
        return vals

    @staticmethod
    def _error_checker_(vals_at_init):
        # Error checking before running procedure
        if np.sum(vals_at_init) is None:
            raise ValueError("When evaluating the estimating equation, `None` was returned. Please check that the "
                             "stacked_equations returns an array evaluated at theta.")
        if np.isnan(np.sum(vals_at_init)):         # Check to see if any np.nan's occur with the initial values
            # Identifying the bad columns
            nans_in_column = np.sum(np.isnan(vals_at_init), axis=0)        # Counting up all NAN's per estimating eq
            columns_w_nans = np.argwhere(nans_in_column >= 1).flatten()    # Returning indices that have any NAN's
            raise ValueError("When evaluated at the initial values, the `stacked_equations` return at least one "
                             "np.nan at the following estimating equation indices: " +
                             str(list(columns_w_nans)) + ". "
                             "As delicatessen does not natively handle missing data, please ensure the "
                             "provided estimating equations resolve any np.nan values accordingly. For details on "
                             "how to handle np.nan's see the documentation at: "
                             "https://deli.readthedocs.io/en/latest/Custom-EE.html#Handling-np.nan")


class MEstimator(_GeneralEstimator):
    r"""M-Estimator for stacked estimating equations.

    Estimating equations are a general approach to point and variance estimation that consists of defining an estimator
    as the solution to a vector of equations that are equal to zero. The corresponding estimators, often called
    M-estimators or Z-estimators, satisify the following equation

    .. math::

        \sum_{i=1}^n \psi(O_i, \hat{\theta}) = 0

    where :math:`\psi` is the :math:`v`-dimensional vector of estimating function(s), :math:`\hat{\theta}` is the
    :math:`v`-dimensional parameter vector, and :math:`O_i` is the observed data (where units are independent but not
    necessarily identically distributed).

    M-Estimators consists of two broad step: point estimation and variance estimation. Point estimation is carried out
    by determining the values of :math:`\theta` where the sum of the estimating equations are zero. This is done via
    standard root-finding algorithms.

    For variance estimation, sandwich variance estimator is used. The asymptotic sandwich variance estimator consists of

    .. math::

        V_n(O, \hat{\theta}) = B_n(O, \hat{\theta})^{-1} F_n(O, \hat{\theta}) \left\{B_n(O, \hat{\theta}^{-1})\right\}^T

    where :math:`B` is the 'bread' and :math:`F` is the 'filling' matrix. These matrices are defined as

    .. math::

        B_n(O, \hat{\theta}) = n^{-1} \sum_{i=1}^{n} - \frac{\partial}{\partial \theta} \psi(O_i, \hat{\theta})

    .. math::

        F_n(O, \hat{\theta}) = n^{-1} \sum_{i=1}^{n} \psi(O_i, \hat{\theta}) \psi(O_i, \hat{\theta})^T

    respectively. The partial derivatives for the bread are calculated using either numerical approximation (e.g.,
    forward difference method) or forward-mode automatic differentiation. Inverting the bread is done via NumPy's
    ``linalg.pinv``. For the filling, the dot product is taken at :math:`\hat{\theta}`.

    After completion of these steps, point and variance estimates are stored. These can be extracted from
    ``MEstimator``. Further, confidence intervals, Z-scores, P-values, or S-values can all be generated.

    Parameters
    ----------
    stacked_equations : function, callable
        Function that returns a `v`-by-`n` NumPy array of the estimating equations. See provided examples in the
        documentation for how to construct a set of estimating equations.
    init : list, set, array
        Initial values for the root-finding algorithm. A total of `v` values should be provided.
    subset : list, set, array, None, optional
        Optional argument to conduct the root-finding procedure on a subset of parameters in the estimating equations.
        The input list is used to location index the parameter array via ``np.take()``. The subset list will
        only affect the root-finding procedure (i.e., the sandwich variance estimator ignores the subset argument).
        Default is ``None``, which runs the root-finding procedure for all parameters in the estimating equations.

    Note
    ----
    Because the root-finding procedure is NOT ran for parameters outside of the subset, those coefficients *must* be
    solved outside of ``MEstimator``. In general, I do *NOT* recommend using the ``subset`` argument unless a series of
    complex estimating equations need to be solved. In general, this argument does not massively improve speed until
    the estimating equations consist of hundreds of parameters.


    Attributes
    ----------
    theta : ndarray
        Parameters for estimating equations. Becomes the estimated parameters after `.estimate()` is called
    variance : ndarray
        Covariance matrix for the parameters, in the same order as the parameter vector
    asymptotic_variance : ndarray
        Asymptotic covariance matrix for the parameters. Not scaled by :math:`n`
    bread : ndarray
        Bread matrix for the parameter vector
    meat : ndarray
        Meat matrix for the parameter vector

    Examples
    --------
    Loading necessary functions and building a generic data set for estimation of the mean

    >>> import numpy as np
    >>> import pandas as pd
    >>> from scipy.stats import logistic
    >>> from delicatessen import MEstimator
    >>> from delicatessen.estimating_equations import ee_mean_variance

    >>> y_dat = [1, 2, 4, 1, 2, 3, 1, 5, 2]

    M-estimation with built-in estimating equation for the mean and variance. First, ``psi``, or the stacked estimating
    equations, is defined

    >>> def psi(theta):
    >>>     return ee_mean_variance(theta=theta, y=y_dat)

    Calling the M-estimation procedure

    >>> estr = MEstimator(stacked_equations=psi, init=[0, 0, ])
    >>> estr.estimate()

    Inspecting the output results

    >>> estr.theta                                  # Point estimates
    >>> estr.variance                               # Covariance
    >>> estr.asymptotic_variance                    # Asymptotic covariance
    >>> np.sqrt(np.diag(estr.asymptotic_variance))  # Standard deviation
    >>> estr.variance                               # Covariance
    >>> np.sqrt(np.diag(estr.variance))             # Standard error
    >>> estr.confidence_intervals()                 # Confidence intervals
    >>> estr.z_scores()                             # Z-scores
    >>> estr.p_values()                             # P-values
    >>> estr.s_values()                             # S-values

    Alternatively, a custom estimating equation can be specified. This is done by constructing a valid estimating
    equation for the ``MEstimator``. The ``MEstimator`` expects the ``psi`` function to return a `v`-by-`n` array,
    where `v` is the number of parameters (length of ``theta``) and n is the total number of observations. Below is an
    example of the mean and variance estimating equation from before, but implemented by-hand

    >>> def psi(theta):
    >>>     y = np.array(y_dat)
    >>>     mean = y - theta[0]
    >>>     variance = (y - theta[0]) ** 2 - theta[1]
    >>>     return mean, variance

    The M-estimation procedure is called using the same approach as before

    >>> estr = MEstimator(stacked_equations=psi, init=[0, 0, ])
    >>> estr.estimate()

    Note that ``len(init)`` should be equal to `v`. So in this case, two initial values are provided.

    ``MEstimator`` can also be run with a user-provided root-finding algorithm. To specify a custom root-finder, a
    function must be created by the user that consists of two keyword arguments (``stacked_equations``, ``init``) and
    must return only the optimized values. The following is an example with SciPy's Levenberg-Marquardt algorithm
    implemented in ``root``.

    >>> def custom_solver(stacked_equations, init):
    >>>     options = {"maxiter": 1000}
    >>>     opt = root(stacked_equations,
    >>>                x0=np.asarray(init),
    >>>                method='lm', tol=1e-9, options=options)
    >>>     return opt.x

    The provided custom root-finder can then be implemented like the following (continuing with the estimating equation
    from the previous example):

    >>> estr = MEstimator(stacked_equations=psi, init=[0, 0, ])
    >>> estr.estimate(solver=custom_solver)

    For more examples on how to apply ``MEstimator``, see https://deli.readthedocs.io/en/latest/

    References
    ----------
    Boos DD, & Stefanski LA. (2013). M-estimation (estimating equations). In Essential Statistical Inference
    (pp. 297-337). Springer, New York, NY.

    Ross RK, Zivich PN, Stringer JSA, & Cole SR. (2024). M-estimation for common epidemiological measures: introduction
    and applied examples. *International Journal of Epidemiology*, 53(2), dyae030.

    Stefanski LA, & Boos DD. (2002). The calculus of M-estimation. *The American Statistician*, 56(1), 29-38.
    """
    def estimate(self, solver='lm', maxiter=5000, tolerance=1e-9, deriv_method='approx', dx=1e-9, allow_pinv=True):
        """Run the point and variance estimation procedures for given estimating equation and starting values. This
        function carries out the point and variance estimation of ``theta``. After this procedure, the point estimates
        (in ``theta``) and the covariance matrix (in ``variance``) can be extracted from the ``MEstimator`` object.

        Parameters
        ----------
        solver : str, function, callable, optional
            Method to use for the root-finding procedure. Default is the Levenberg-Marquardt algorithm
            (``scipy.optimize.root(method='lm')``, specified by ``solver='lm'``). Other built-in option is the secant
            method (``scipy.optimize.newton``, specified by ``solver='newton'``), and a modification of the Powell
            hybrid method (``scipy.optimize.root(method='hybr')``, specified by ``solver='hybr'``). Finally, any generic
            root-finding algorithm can be used via a user-provided callable object. The function must consist of two
            keyword arguments: ``stacked_equations``, and ``init``. Additionally, the function should return only the
            optimized values. Please review the provided example in the documentation for how to implement a custom
            root-finding algorithm.
        maxiter : int, optional
            Maximum iterations to consider for the root finding procedure. Default is 5000 iterations. For complex
            estimating equations, this value may need to be increased. This argument is not used when a custom
            root-finding method (e.g., ``solver``) is provided.
        tolerance : float, optional
            Maximum tolerance for errors in the root finding in ``scipy.optimize``. Default is 1e-9. This argument is
            not used when a custom root-finding method (e.g., ``solver``) is provided.
        deriv_method : str, optional
            Method to compute the derivative of the estimating equations for the bread matrix. Default is ``'approx'``.
            Options include numerical approximation via the forward difference method via SciPy (``'approx'``), forward
            difference as implemented in delicatessen (`'fapprox'`), backward difference as implemented in delicatessen
            (`'bapprox'`), central difference implemented as in delicatessen (`'capprox'`), or forward-mode automatic
            differentiation as implemented in delicatessen(``'exact'``).
        dx : float, optional
            Spacing to use to numerically approximate the partial derivatives of the bread matrix. Default is 1e-9.
            Here, a small value for ``dx`` should be used, since some large values can result in poor approximations.
            This argument is only used with numerical approximation methods.
        allow_pinv : bool, optional
            Whether to allow for the pseudo-inverse (via ``numpy.linalg.pinv``) if the bread matrix is determined to be
            non-invertible. If you want to disallow the pseudo-inverse (i.e., use ``numpy.linalg.inv``), set this
            argument to ``False``. Default is ``True``, which  is more robust to the possible bread matrices.

        Returns
        -------
        None
        """
        # Evaluate stacked estimating equations at init
        vals_at_init = self.stacked_equations(theta=self.init)    # Calculating the initial values
        vals_at_init = np.asarray(vals_at_init                    # Convert output to an array (in case it isn't)
                                  ).T                             # ... transpose so N is always the 1st element

        # Error checking for inputs
        self._error_checker_(vals_at_init=vals_at_init)

        # Checking dimensions of output for M-estimator
        if vals_at_init.ndim == 1 and np.asarray(self.init).shape[0] == 1:     # Checks to ensure dimensions align
            # the starting if-state is to work-around inits=[0, ] (otherwise breaks the first else-if)
            pass
        elif vals_at_init.ndim == 1 and np.asarray(self.init).shape[0] != 1:
            raise ValueError("The number of initial values and the number of rows returned by `stacked_equations` "
                             "should be equal but there are " + str(np.asarray(self.init).shape[0]) + " initial values "
                             "and the `stacked_equations` function returns " + str(1) + " row.")
        elif np.asarray(self.init).shape[0] != vals_at_init.shape[1]:
            raise ValueError("The number of initial values and the number of rows returned by `stacked_equations` "
                             "should be equal but there are " + str(np.asarray(self.init).shape[0]) + " initial values "
                             "and the `stacked_equations` function returns " + str(vals_at_init.shape[1])
                             + " row(s).")
        elif vals_at_init.ndim > 2:
            raise ValueError("A 2-dimensional array is expected, but the `stacked_equations` returns a "
                             + str(vals_at_init.ndim) + "-dimensional array.")
        else:
            pass

        # Trick to get the number of observations from the estimating equations (transposed above)
        self.n_obs = vals_at_init.shape[0]

        # STEP 1: solving the M-estimator stacked equations
        # To allow for optimization of only a subset of parameters in the estimating equation (in theory meant to
        #   simplify the process of complex stacked estimating equations where pre-washing can be done effectively),
        #   we do some internal processing. Essentially, we 'freeze' the parameters outside of self._subset_ as their
        #   inits, and let the root-finding procedure update the self._subset_ parameters. We do this by subsetting out
        #   the init values then passing them along to root(). Behind the scenes, self._mestimation_answer_() expands
        #   the parameters (to include everything), calculates the estimating equation at those values, and then
        #   extracts the corresponding subset.
        #   This process only takes place within Step 1. There is an inherent danger with this process in that if
        #   non-subset parameters are not pre-washed, then the returned parameters will not be correct. I am
        #   considering adding a warning for self_subset_, but I currently just trust the user...

        # Processing initial values based on whether subset option was specified
        if self._subset_ is None:                        # If NOT subset,
            inits = self.init                            # ... give all initial values
        else:                                            # If subset,
            inits = np.take(self.init, self._subset_)    # ... then extract initial values for the subset

        # Calculating parameters values via the root-finder (for only the subset of values!)
        slv_theta = self._solve_coefficients_(stacked_equations=self._mestimation_answer_,  # Give the EE's
                                              init=inits,                                   # Give the subset vals
                                              method=solver,                                # Specify the solver
                                              maxiter=maxiter,                              # Set max iterations
                                              tolerance=tolerance                           # Set allowable tolerance
                                              )

        # Processing parameters after the root-finding procedure
        if self._subset_ is None:                        # If NOT subset,
            self.theta = slv_theta                       # ... then use the full output/solved theta
        else:                                            # If subset,
            self.theta = np.asarray(self.init)           # ... copy the initial values
            for s, n in zip(self._subset_, slv_theta):   # ... then look over the subset and input theta
                self.theta[s] = n                        # ... and update the subset to the output/solved theta

        # STEP 2: calculating the sandwich variance
        # After solving for the parameters, we now can compute the empirical sandwich variance estimator. This is
        #   done by compute the bread and meat matrices and then combining them. This is now done by a separate
        #   functionalities within the `sandwich.py` file as of v2.2.
        # STEP 2.1: baking the Bread
        self.bread = compute_bread(stacked_equations=self.stacked_equations,
                                   theta=self.theta,
                                   deriv_method=deriv_method,
                                   dx=dx) / self.n_obs

        # STEP 2.2: slicing the meat
        self.meat = compute_meat(stacked_equations=self.stacked_equations,
                                 theta=self.theta) / self.n_obs

        # STEP 2.3: assembling the sandwich (variance)
        self.asymptotic_variance = build_sandwich(bread=self.bread,
                                                  meat=self.meat,
                                                  allow_pinv=allow_pinv)
        if self.asymptotic_variance is None:
            self.variance = self.asymptotic_variance
        else:
            self.variance = self.asymptotic_variance / self.n_obs

    def _mestimation_answer_(self, theta):
        """Internal function to evaluate the sum of the estimating equations. The summation is internally evaluated
        since access to the estimating functions is needed for the sandwich variance computations. This function is
        used by the root-finding procedure (since we need the subset applied).

        Parameters
        ----------
        theta : array
            b-by-n matrix to sum over the values of n.

        Returns
        -------
        array :
            b-by-1 array, which is the sum over n for each b.
        """
        # Option for the subset argument
        if self._subset_ is None:                      # If NOT subset then,
            full_theta = theta                         # ... then use the full input theta
        else:                                          # If subset then,
            full_theta = np.asarray(self.init)         # ... copy the initial values to ndarray
            np.put(full_theta,                         # ... update in place the previous array
                   ind=self._subset_,                  # ... go to the subset indices
                   v=theta)                            # ... then input current iteration values

        stacked_equations = np.asarray(self.stacked_equations(full_theta))  # Returning stacked equation
        return self._eval_ee_(stacked_equations=stacked_equations,   # Passing to evaluating function
                              subset=self._subset_)                  # ... with specified subset

    @staticmethod
    def _solve_coefficients_(stacked_equations, init, method, maxiter, tolerance):
        """Calls the root-finding procedure for the values of theta, such that the estimating equations are equal to
        zero. Default uses the Levenberg-Marquardt algorithm from SciPy.

        Parameters
        ----------
        stacked_equations : function
            Function that contains the estimating equations
        init : array
            Initial values for the optimizer
        method : str, function, callable
            Method to use for the root finding procedure. Can be either a string or a callable object
        maxiter : int
            Maximum iterations to consider for the root finding procedure
        tolerance : float
            Maximum tolerance for errors in the root finding. This is SciPy's `tol` argument.

        Returns
        -------
        psi: array
            Solved or optimal values for theta given the estimating equations and data
        """
        # Newton solver has a special catch (is not nested in root())
        if method == "newton":
            psi = newton(stacked_equations,    # ... stacked equations to solve (should be written as sums)
                         x0=np.asarray(init),  # ... initial values for solver
                         # full_output=True,   # ... returns full output (rather than just the root)
                         maxiter=maxiter,      # ... setting maximum number of iterations
                         tol=tolerance,        # ... setting some tolerance values (should be able to update)
                         disp=True)            # ... option to raise RuntimeError if doesn't converge
        # Otherwise, goes to root with some selected optimizers
        elif method in ['lm', ]:
            options = {"maxiter": maxiter}
            opt = root(stacked_equations,      # ... stacked equations to solve (should be written as sums)
                       x0=np.asarray(init),    # ... initial values for solver
                       method=method,          # ... allow for valid root-finding methods
                       tol=tolerance,          # ... setting some tolerance values
                       options=options)        # ... options for the selected solver
            psi = opt.x                        # Error handling if fails to converge
            if opt.success == 0:
                print("Root-finding failed to converge...")
                raise RuntimeError(opt.message)
        elif method in ['hybr', ]:
            options = {"maxfev": maxiter}
            opt = root(stacked_equations,      # ... stacked equations to solve (should be written as sums)
                       x0=np.asarray(init),    # ... initial values for solver
                       method=method,          # ... allow for valid root-finding methods
                       tol=tolerance,          # ... setting some tolerance values
                       options=options)        # ... allow for options in hybrid
            psi = opt.x                        # Error handling if fails to converge
            if opt.success == 0:
                print("Root-finding failed to converge...")
                raise RuntimeError(opt.message)
        elif callable(method):
            try:
                psi = method(stacked_equations=stacked_equations,
                             init=np.asarray(init))
            except TypeError:
                raise TypeError("The user-specified root-finding `solver` must be a function (or callable object) with "
                                "the following keyword arguments: `stacked_equations`, `init`.")
            if psi is None:
                raise ValueError("The user-specified root-finding `solver` must return the solution to the "
                                 "optimization")
        else:
            # ... otherwise throw ValueError if no other root-finding steps are triggered.
            raise ValueError("The solver '" + str(method) + "' is not available. Please see the "
                             "documentation for valid options for the optimizer.")

        # Return optimized theta array
        return psi


class GMMEstimator(_GeneralEstimator):
    r"""Generalized Method of Moments (GMM) Estimator for stacked estimating equations.

    Estimating equations are a general approach to point and variance estimation that consists of defining an estimator
    as the solution to a vector of equations that are equal to zero. The corresponding GMM estimator satisifies the
    following equation

    .. math::

        \text{argmin}_{\theta} \left[ \sum_{i=1}^n \psi(O_i, \hat{\theta}) \right]^T
        \text{Q}
        \left[ \sum_{i=1}^n \psi(O_i, \hat{\theta}) \right]^T

    where :math:`\psi` is the :math:`\ge v`-dimensional vector of estimating function(s), :math:`\hat{\theta}` is the
    :math:`v`-dimensional parameter vector, :math:`\text{Q}` is a weight matrix, and :math:`O_i` is the observed data
    (where units are independent but not necessarily identically distributed).

    GMM estimators consists of two broad step: point estimation and variance estimation. Point estimation is carried
    out by finding where the minimum of the negative product of the estimating equations is. Initially :math:`\text{Q}`
    is set to be the identity matrix. This process is done via standard minimization algorithms.

    Note
    ----
    For over-identified settings, ``GMMEstimator`` uses an two-step iterative procedure, where :math:`\theta` is
    estimated, then the meat matrix is computed. The meat matrix replaces the identity matrix as :math:`\text{Q}`. Then
    the optimization process is repeated.


    For variance estimation, sandwich variance estimator is used. The asymptotic sandwich variance estimator consists of

    .. math::

        V_n(O, \hat{\theta}) = B_n(O, \hat{\theta})^{-1} F_n(O, \hat{\theta}) \left\{B_n(O, \hat{\theta}^{-1})\right\}^T

    where :math:`B` is the 'bread' and :math:`F` is the 'filling' matrix. These matrices are defined as

    .. math::

        B_n(O, \hat{\theta}) = n^{-1} \sum_{i=1}^{n} - \frac{\partial}{\partial \theta} \psi(O_i, \hat{\theta})

    .. math::

        F_n(O, \hat{\theta}) = n^{-1} \sum_{i=1}^{n} \psi(O_i, \hat{\theta}) \psi(O_i, \hat{\theta})^T

    respectively. The partial derivatives for the bread are calculated using either numerical approximation (e.g.,
    forward difference method) or forward-mode automatic differentiation. Inverting the bread is done via NumPy's
    ``linalg.pinv``. For the filling, the dot product is taken at :math:`\hat{\theta}`.

    After completion of these steps, point and variance estimates are stored. These can be extracted from
    ``GMMEstimator``. Further, confidence intervals, Z-scores, P-values, or S-values can all be generated.

    Parameters
    ----------
    stacked_equations : function, callable
        Function that returns a `v`-by-`n` NumPy array of the estimating equations. See provided examples in the
        documentation for how to construct a set of estimating equations.
    init : list, set, array
        Initial values for the minimization algorithm. A total of `v` values should be provided.
    subset : list, set, array, None, optional
        Optional argument to conduct the minimization procedure on a subset of parameters in the estimating equations.
        The input list is used to location index the parameter array via ``np.take()``. The subset list will
        only affect the minimization procedure (i.e., the sandwich variance estimator ignores the subset argument).
        Default is ``None``, which runs the minimization procedure for all parameters in the estimating equations.
    overid_maxiter : int, optional
        Maximum number of iterations for the two-step GMM-estimation procedure with over-identified parameters. Note
        that this procedure is only used for estimating equations that include an over-identified parameter, otherwise
        this argument has no impact. Default is ``10``.
    overid_tolerance : float, optional
        Error tolerance for the two-step GMM-estimation procedure with over-identified parameters. Note that this
        procedure is only used for estimating equations that include an over-identified parameter, otherwise this
        argument has no impact. If the tolerance threshold is met, the two-step procedure is terminated. The threshold
        is defined as the maximum of the absolute value of the difference between the parameters of the current
        iteration and the previous iteration. Default is ``1e-9``.

    Note
    ----
    Because the minimization procedure is NOT ran for parameters outside of the subset, those coefficients *must* be
    solved outside of ``GMMEstimator``. In general, I do *NOT* recommend using the ``subset`` argument unless a series
    of complex estimating equations need to be solved. In general, this argument does not massively improve speed until
    the estimating equations consist of hundreds of parameters.

    Attributes
    ----------
    theta : ndarray
        Parameters for estimating equations. Becomes the estimated parameters after `.estimate()` is called
    variance : ndarray
        Covariance matrix for the parameters, in the same order as the parameter vector
    asymptotic_variance : ndarray
        Asymptotic covariance matrix for the parameters. Not scaled by :math:`n`
    bread : ndarray
        Bread matrix for the parameter vector
    meat : ndarray
        Meat matrix for the parameter vector
    weight_matrix : ndarray
        Weight matrix, :math:`\text{Q}` used. For just-identified problems, the weight matrix is simply the identity
        matrix, otherwise the weight matrix is updated to the meat matrix from the prior iteration.

    Examples
    --------
    Loading necessary functions and building a generic data set for estimation of the mean

    >>> import numpy as np
    >>> import pandas as pd
    >>> from scipy.stats import logistic
    >>> from delicatessen import GMMEstimator
    >>> from delicatessen.estimating_equations import ee_mean_variance

    >>> y_dat = [1, 2, 4, 1, 2, 3, 1, 5, 2]

    GMM-estimation with built-in estimating equation for the mean and variance. First, ``psi``, or the stacked
    estimating equations, is defined

    >>> def psi(theta):
    >>>     return ee_mean_variance(theta=theta, y=y_dat)

    Calling the GMM-estimation procedure

    >>> estr = GMMEstimator(stacked_equations=psi, init=[0, 0, ])
    >>> estr.estimate()

    Inspecting the output results

    >>> estr.theta                                  # Point estimates
    >>> estr.variance                               # Covariance
    >>> estr.asymptotic_variance                    # Asymptotic covariance
    >>> np.sqrt(np.diag(estr.asymptotic_variance))  # Standard deviation
    >>> estr.variance                               # Covariance
    >>> np.sqrt(np.diag(estr.variance))             # Standard error
    >>> estr.confidence_intervals()                 # Confidence intervals
    >>> estr.z_scores()                             # Z-scores
    >>> estr.p_values()                             # P-values
    >>> estr.s_values()                             # S-values

    For more examples on how to apply ``GMMEstimator``, see https://deli.readthedocs.io/en/latest/

    References
    ----------
    Duncan GM. (1987). A simplified approach to M-estimation with application to two-stage estimators.
    *Journal of Econometrics*, 34(3), 373-389.

    Hansen LP. (1982). Large sample properties of generalized method of moments estimators.
    *Econometrica: Journal of the Econometric Society*, 1029-1054.

    Jesus J, & Chandler RE. (2011). Estimating functions and the generalized method of moments. *Interface Focus*,
    1(6), 871-885.

    Wirjanto T. (1997). Estimating Functions and Over-Identified Models. *Lecture Notes - Monograph Series*, 239-256.
    """
    def __init__(self, stacked_equations, init, subset=None, overid_maxiter=10, overid_tolerance=1e-9):
        # Calling BaseClass for the initial setup
        super().__init__(stacked_equations=stacked_equations, init=init, subset=subset)

        # Processing GMM-specific arguments
        self._over_id_iterations_ = overid_maxiter
        self._over_id_tolerance_ = overid_tolerance

    def estimate(self, solver='bfgs', maxiter=5000, tolerance=1e-9, deriv_method='approx', dx=1e-9, allow_pinv=True):
        """Run the point and variance estimation procedures for given estimating equation and starting values. This
        function carries out the point and variance estimation of ``theta``. After this procedure, the point estimates
        (in ``theta``) and the covariance matrix (in ``variance``) can be extracted from the ``GMMEstimator`` object.

        Parameters
        ----------
        solver : str, function, callable, optional
            Method to use for the minimization procedure. Default is the BFGS algorithm
            (``scipy.optimize.minimize(method='bfgs')``, specified by ``solver='bfgs'``). Other built-in options are
            the L-BFGS-B algorithm (``'l-bfgs-b'``),
            the conjugate gradient algorithm (``'cg'``),
            Nelder-Mead (``'nelder-mead'``),
            and the modified Powell algorithm (``'powell'``).
            Finally, any generic minimization algorithm can be used via a user-provided callable object. The function
            must consist of two keyword arguments: ``stacked_equations``, and ``init``. Additionally, the function
            should return only the optimized values. Please review the provided example in the documentation for how to
            implement a custom minimization algorithm.
        maxiter : int, optional
            Maximum iterations to consider for the minimization procedure. Default is 5000 iterations. For complex
            estimating equations, this value may need to be increased. This argument is not used when a custom
            minimization method (e.g., ``solver``) is provided.
        tolerance : float, optional
            Maximum tolerance for errors in the minimization in ``scipy.optimize``. Default is 1e-9. This argument is
            not used when a custom minimization method (e.g., ``solver``) is provided.
        deriv_method : str, optional
            Method to compute the derivative of the estimating equations for the bread matrix. Default is ``'approx'``.
            Options include numerical approximation via the forward difference method via SciPy (``'approx'``), forward
            difference as implemented in delicatessen (`'fapprox'`), backward difference as implemented in delicatessen
            (`'bapprox'`), central difference implemented as in delicatessen (`'capprox'`), or forward-mode automatic
            differentiation as implemented in delicatessen(``'exact'``).
        dx : float, optional
            Spacing to use to numerically approximate the partial derivatives of the bread matrix. Default is 1e-9.
            Here, a small value for ``dx`` should be used, since some large values can result in poor approximations.
            This argument is only used with numerical approximation methods.
        allow_pinv : bool, optional
            Whether to allow for the pseudo-inverse (via ``numpy.linalg.pinv``) if the bread matrix is determined to be
            non-invertible. If you want to disallow the pseudo-inverse (i.e., use ``numpy.linalg.inv``), set this
            argument to ``False``. Default is ``True``, which  is more robust to the possible bread matrices.

        Returns
        -------
        None
        """
        # Evaluate stacked estimating equations at init
        vals_at_init = self.stacked_equations(theta=self.init)    # Calculating the initial values
        vals_at_init = np.asarray(vals_at_init                    # Convert output to an array (in case it isn't)
                                  ).T                             # ... transpose so N is always the 1st element

        # Error checking for inputs
        self._error_checker_(vals_at_init=vals_at_init)

        # Processing dependent on number of estimating functions
        if vals_at_init.ndim == 1:
            n_params = 1
            over_identified = False
            if np.asarray(self.init).shape[0] > 1:
                raise ValueError(
                    "The number of initial values should be less than or equal to the number of rows "
                    "returned by `stacked_equations` but there are " + str(np.asarray(self.init).shape[0])
                    + " initial values and the `stacked_equations` function returns 1 row.")
        # If estimating functions are of 2 dimensions, need to check for over-identification
        elif vals_at_init.ndim == 2:
            n_params = vals_at_init.shape[1]                        # Getting the number of parameters
            if np.asarray(self.init).shape[0] == n_params:          # Finding if the problem is just-identified
                over_identified = False                             # ... flag for over-ID iteration procedure
            elif np.asarray(self.init).shape[0] < n_params:         # Finding if the problem is over-identified
                over_identified = True                              # ... flag for over-ID iteration procedure
            else:                                                   # Under-identified parameters are not supported
                raise ValueError(
                    "The number of initial values should be less than or equal to the number of rows "
                    "returned by `stacked_equations` but there are " + str(np.asarray(self.init).shape[0])
                    + " initial values and the `stacked_equations` function returns "
                    + str(vals_at_init.shape[1]) + " row(s).")

        # Catching if user provided invalid estimating function shape
        elif vals_at_init.ndim > 2:
            raise ValueError("A 2-dimensional array is expected, but the `stacked_equations` returns a "
                             + str(vals_at_init.ndim) + "-dimensional array.")

        # Nothing should arrive here via logic of previous operations
        else:
            raise ValueError("This ValueError should not be reached by the structure of the problem. If you receive "
                             "this error, please reach out https://github.com/pzivich/Delicatessen")

        # Trick to get the number of observations from the estimating equations (transposed above)
        self.n_obs = vals_at_init.shape[0]

        # Obtaining initial weight matrix
        self.weight_matrix = np.identity(n=n_params)

        # Updating theta to init values for minimization call (since needed in _gmmestimation_answer_)
        self.theta = self.init

        # STEP 1: solving the GMM-estimator stacked equations
        # To allow for optimization of only a subset of parameters in the estimating equation (in theory meant to
        #   simplify the process of complex stacked estimating equations where pre-washing can be done effectively),
        #   we do some internal processing. Essentially, we 'freeze' the parameters outside of self._subset_ as their
        #   inits, and let the minimization procedure update the self._subset_ parameters. We do this by subsetting out
        #   the init values then passing them along to solver. Behind the scenes, self._solve_coefficients_() expands
        #   the parameters (to include everything), calculates the estimating equation at those values, and then
        #   extracts the corresponding subset.
        #   This process only takes place within Step 1. There is an inherent danger with this process in that if
        #   non-subset parameters are not pre-washed, then the returned parameters will not be correct. I am
        #   considering adding a warning for self_subset_, but I currently just trust the user...
        # STEP 1.1: Initial solving procedure for point estimation
        self._solve_for_theta_(solver=solver, maxiter=maxiter, tolerance=tolerance, init=self.theta)

        # STEP 1.2: Iterative procedure if we are in the over-identified setting
        if over_identified:
            # Iterative process argument processing
            current_iteration = 0
            if self._over_id_iterations_ < 1:
                terminate = True
            else:
                terminate = False

            # Applying the iterative procedure
            while not terminate:
                # Setup for current iteration
                current_iteration = current_iteration + 1
                prev_theta = np.asarray(self.theta)
                self.weight_matrix = self._get_q_matrix_(allow_pinv=allow_pinv)
                self._solve_for_theta_(solver=solver, maxiter=maxiter,
                                       tolerance=tolerance, init=self.theta)

                # Checking ending conditions
                error = np.abs(np.asarray(self.theta) - prev_theta)
                if np.max(error) <= self._over_id_tolerance_:
                    terminate = True
                if current_iteration >= self._over_id_iterations_:
                    terminate = True
                    warnings.warn("Maximum number of iterations, `overid_maxiters=" + str(self._over_id_iterations_)
                                  + "`, has been exceeded for the iterative GMM updating. Terminating the "
                                  + "over-identification procedure at the last estimated values.",
                                  UserWarning)

        # STEP 2: calculating the sandwich variance
        # After solving for the parameters, we now can compute the empirical sandwich variance estimator. This is
        #   done by compute the bread and meat matrices and then combining them. This is now done by a separate
        #   functionalities within the `sandwich.py` file as of v2.2.
        # STEP 2.1: baking the Bread
        self.bread = compute_bread(stacked_equations=self.stacked_equations,
                                   theta=self.theta,
                                   deriv_method=deriv_method,
                                   dx=dx) / self.n_obs

        # STEP 2.2: slicing the meat
        self.meat = compute_meat(stacked_equations=self.stacked_equations,
                                 theta=self.theta) / self.n_obs

        # STEP 2.3: assembling the sandwich (variance)
        self.asymptotic_variance = build_sandwich(bread=self.bread,
                                                  meat=self.meat,
                                                  allow_pinv=allow_pinv)
        if self.asymptotic_variance is None:
            self.variance = self.asymptotic_variance
        else:
            self.variance = self.asymptotic_variance / self.n_obs

    def _gmmestimation_answer_(self, theta):
        """Internal function to evaluate the function to be minimized over. The summation is internally evaluated
        since access to the estimating functions is needed for the sandwich variance computations. This function is
        used by the minimization procedure (since we need the subset applied).

        Parameters
        ----------
        theta : array
            b-by-n matrix to sum over the values of n.

        Returns
        -------
        array :
            b-by-1 array, which is the sum over n for each b.
        """
        # Option for the subset argument
        if self._subset_ is None:                      # If NOT subset then,
            full_theta = theta                         # ... then use the full input theta
        else:                                          # If subset then,
            full_theta = np.asarray(self.theta)        # ... copy the initial values to ndarray
            np.put(full_theta,                         # ... update in place the previous array
                   ind=self._subset_,                  # ... go to the subset indices
                   v=theta)                            # ... then input current iteration values

        # Evaluating estimating functions
        stacked_equations = np.asarray(self.stacked_equations(full_theta))  # Returning stacked equation
        est_eqtns = self._eval_ee_(stacked_equations=stacked_equations,     # Passing to evaluating function
                                   subset=None)                             # ... with specified subset
        est_eqtns = est_eqtns / self.n_obs                                  # Scaling EE by n (helpful for minimizer)

        # Evaluating GMM estimator at given theta
        return np.dot(np.dot(est_eqtns, self.weight_matrix), est_eqtns)

    def _solve_for_theta_(self, solver, maxiter, tolerance, init):
        """Internal function to evaluate and solve the estimating equations for theta. This is expressed as an internal
        function to simplify the iterative procedure for over-identified problems

        Parameters
        ----------
        solver : str, function, callable, optional
            Method to use for the minimization procedure.
        maxiter : int, optional
            Maximum iterations to consider for the minimization procedure.
        tolerance : float, optional
            Maximum tolerance for errors.

        Returns
        -------
        None
        """
        # TODO I added init as an argument rather than calling self.init. I want to think about this more
        if self._subset_ is None:                        # If NOT subset,
            inits = init                                 # ... give all initial values
        else:                                            # If subset,
            inits = np.take(init, self._subset_)         # ... then extract initial values for the subset

        # Calculating parameters values via the minimizer (for only the subset of values!)
        slv_theta = self._solve_coefficients_(stacked_equations=self._gmmestimation_answer_,
                                              init=inits,
                                              method=solver,
                                              maxiter=maxiter,
                                              tolerance=tolerance)

        # Processing parameters after the minimization procedure
        if self._subset_ is None:                        # If NOT subset,
            self.theta = slv_theta                       # ... then use the full output/solved theta
        else:                                            # If subset,
            self.theta = np.asarray(self.theta)          # ... copy the initial values
            for s, n in zip(self._subset_, slv_theta):   # ... then look over the subset and input theta
                self.theta[s] = n                        # ... and update the subset to the output/solved theta

    def _get_q_matrix_(self, allow_pinv):
        """Internal function to compute the weight matrix Q for over-identified problems

        Returns
        -------

        """
        # Compute the outer product (meat matrix)
        meat_matrix = compute_meat(stacked_equations=self.stacked_equations, theta=self.theta) / self.n_obs

        # Invert meat matrix
        if allow_pinv:
            q_matrix = np.linalg.pinv(meat_matrix)
        else:
            q_matrix = np.linalg.inv(meat_matrix)
        return q_matrix

    @staticmethod
    def _solve_coefficients_(stacked_equations, init, method, maxiter, tolerance):
        """Calls the minimization procedure for the values of theta, such that the estimating equations are equal to
        zero. Default uses the BFGS algorithm from SciPy.

        Parameters
        ----------
        stacked_equations : function
            Function that contains the estimating equations
        init : array
            Initial values for the optimizer
        method : str, function, callable
            Method to use for the minimization procedure. Can be either a string or a callable object
        maxiter : int
            Maximum iterations to consider for the minimization procedure
        tolerance : float
            Maximum tolerance for errors in the minimization. This is SciPy's `tol` argument.

        Returns
        -------
        psi: array
            Solved or optimal values for theta given the estimating equations and data
        """
        # Default SciPy minimizers provided that are valid to use here
        if method in ['cg', 'bfgs', 'l-bfgs-b', 'powell', 'nelder-mead']:
            opt = minimize(stacked_equations,            # ... stacked equations to solve (should be written as sums)
                           x0=np.asarray(init),          # ... initial values for solver
                           method=method,                # ... chosen minimization method
                           options={"maxiter": maxiter,  # ... setting maximum number of iterations
                                    "disp": False},      # ... don't display results (SciPy prints stuff on success)
                           tol=tolerance)                # ... setting some tolerance values (should be able to update)
            if opt.status != 0:                                # Check whether optimization was successful
                warnings.warn(opt.message, OptimizeWarning)    # ... if not, throw the SciPy standard warning to user
            psi = opt.x                                        # Extract the estimated parameter vector

        # Callable custom minimizer provided by the user
        # TODO still need to see if pytorch works here
        elif callable(method):
            try:
                psi = method(stacked_equations=stacked_equations,
                             init=np.asarray(init))
            except TypeError:
                raise TypeError("The user-specified minimizer `solver` must be a function (or callable object) with "
                                "the following keyword arguments: `stacked_equations`, `init`.")
            if psi is None:
                raise ValueError("The user-specified minimizer `solver` must return the solution to the "
                                 "optimization")

        # Value error for unsupported solvers
        else:
            # ... otherwise throw ValueError if no other minimizer steps are triggered.
            raise ValueError("The solver '" + str(method) + "' is not available. Please see the "
                             "documentation for valid options for the optimizer.")

        # Returning the theta estimates
        return psi
